import dataclasses
import types
from collections.abc import Sequence
from unittest import mock

from django.db import models

from extended_mypy_django_plugin.django_analysis import ImportPath, Project, discovery, protocols


class TestKnownModelsDiscovery:
    def test_it_finds_modules_that_have_models(
        self, loaded_django_example: protocols.Loaded[Project]
    ) -> None:
        @dataclasses.dataclass
        class ExampleModule:
            @classmethod
            def create(
                cls,
                *,
                import_path: protocols.ImportPath,
                module: types.ModuleType | None,
                models: Sequence[type[models.Model]],
            ) -> protocols.Module:
                defined_models = discovery.make_module_creator()(
                    import_path=import_path, module=module, models=models
                ).defined_models
                return cls(
                    installed=module is not None,
                    import_path=import_path,
                    module=module,
                    models=models,
                    defined_models=defined_models,
                )

            installed: bool
            module: types.ModuleType | None
            models: Sequence[type[models.Model]]
            import_path: protocols.ImportPath
            defined_models: protocols.ModelMap

        known_modules = discovery.DefaultInstalledModulesDiscovery[Project](
            module_creator=ExampleModule.create
        )(loaded_django_example)

        import django.contrib.admin.models
        import django.contrib.auth.base_user
        import django.contrib.auth.models
        import django.contrib.contenttypes.models
        import django.contrib.sessions.base_session
        import django.contrib.sessions.models
        import djangoexample.empty_models.models
        import djangoexample.exampleapp.models
        import djangoexample.exampleapp2.models
        import djangoexample.only_abstract.models
        import djangoexample.relations1.models
        import djangoexample.relations2.models

        expected = [
            ExampleModule(
                installed=True,
                import_path=ImportPath("django.contrib.admin.models"),
                module=(mod := django.contrib.admin.models),
                models=[mod.LogEntry],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("django.contrib.auth.models"),
                module=(mod := django.contrib.auth.models),
                models=[
                    mod.AbstractUser,
                    mod.PermissionsMixin,
                    mod.Permission,
                    mod.Group,
                    mod.User,
                ],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("django.contrib.contenttypes.models"),
                module=(mod := django.contrib.contenttypes.models),
                models=[mod.ContentType],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("django.contrib.sessions.models"),
                module=(mod := django.contrib.sessions.models),
                models=[mod.Session],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("djangoexample.exampleapp.models"),
                module=(mod := djangoexample.exampleapp.models),
                models=[mod.Parent, mod.Parent2, mod.Child1, mod.Child2, mod.Child3, mod.Child4],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("djangoexample.exampleapp2.models"),
                module=(mod := djangoexample.exampleapp2.models),
                models=[mod.ChildOther, mod.ChildOther2],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("djangoexample.only_abstract.models"),
                module=(mod := djangoexample.only_abstract.models),
                models=[mod.AnAbstract],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("djangoexample.relations1.models"),
                module=(mod := djangoexample.relations1.models),
                models=[mod.Abstract, mod.Child1, mod.Child2, mod.Concrete1, mod.Concrete2],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("djangoexample.relations2.models"),
                module=(mod := djangoexample.relations2.models),
                models=[mod.Thing],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("djangoexample.empty_models.models"),
                module=(mod := djangoexample.empty_models.models),
                models=[],
                defined_models={},
            ),
            # # There is no way for our discovery to know about these without some kind of invasive
            # # traversal of every file in site-packages. We do know about them at mypy time and create
            # # virtual dependencies for them when they are seen. We also can't import them without them
            # # being installed because you can't create a model class on an app that is not installed
            # # So any code that depends on the annotations that would go into the virtual dependencies
            # # cannot import the models the annotations would be used with so this is fine
            # Module(
            #     installed=False,
            #     import_path=ImportPath("djangoexample.not_installed_only_abstract.models"),
            #     module=None,
            #     models=[],
            #     defined_models=mock.ANY,
            # ),
            # Module(
            #     installed=False,
            #     import_path=ImportPath("djangoexample.not_installed_with_concrete.models"),
            #     module=None,
            #     models=[],
            #     defined_models=mock.ANY,
            # ),
            ###
            ## Indirect modules
            ###
            ExampleModule(
                installed=True,
                import_path=ImportPath("django.contrib.auth.base_user"),
                module=(mod := django.contrib.auth.base_user),
                models=[mod.AbstractBaseUser],
                defined_models=mock.ANY,
            ),
            ExampleModule(
                installed=True,
                import_path=ImportPath("django.contrib.sessions.base_session"),
                module=(mod := django.contrib.sessions.base_session),
                models=[mod.AbstractBaseSession],
                defined_models=mock.ANY,
            ),
        ]
        assert known_modules == {m.import_path: m for m in expected}



def interface____differentiated__9() -> None:
    return None

mod = "djangoexample.only_abstract.models"
summary = "__virtual__.mod_4035906997::djangoexample.only_abstract.models::installed_apps=__installed_apps_hash__::significant=1262772787"

import djangoexample.only_abstract.models



import ast
import dataclasses
import importlib
import inspect
import pathlib
import subprocess
import sys
import tempfile
import textwrap
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, TypeVar, cast

from pytest_typing_runner import builders, file_changers, protocols, scenarios
from typing_extensions import Self

scripts_dir = pathlib.Path(__file__).parent.parent.parent

T_Scenario = TypeVar("T_Scenario", bound="Scenario")


@dataclasses.dataclass(kw_only=True)
class ScenarioInfo:
    django_settings_module: str = "mysettings"
    mypy_configuration_filename: str = "mypy.ini"
    mypy_configuration_content: str = """
        [mypy]
        mypy_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/test

        plugins =
            extended_mypy_django_plugin.main

        [mypy.plugins.django-stubs]
        scratch_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/test
        django_settings_module = {django_settings_module}
        """
    additional_mypy_configuration_content: str = ""
    copied_apps: list[str] = dataclasses.field(default_factory=list)
    installed_apps: list[str] = dataclasses.field(default_factory=list)
    monkeypatch: bool = False
    debug: bool = False
    change_settings: Callable[[str], str] | None = None


@dataclasses.dataclass(frozen=True, kw_only=True)
class Scenario(scenarios.Scenario):
    info: ScenarioInfo = dataclasses.field(default_factory=ScenarioInfo)

    def copy_django_app(
        self,
        app: str,
        file_modification: protocols.FileModifier,
        skip_if_destination_exists: bool = True,
    ) -> None:
        location = scripts_dir / app
        if location.exists() and location.is_dir() and location.is_relative_to(scripts_dir):
            file_changers.CopyDirectory(root_dir=self.root_dir, src=scripts_dir, path=app).do_copy(
                modify_file=file_modification,
                exclude=lambda p: p.suffix == ".pyc",
                skip_if_destination_exists=skip_if_destination_exists,
            )

    def determine_django_settings_content(
        self, settings_path: str, options: protocols.RunOptions[Self]
    ) -> str:
        found = {}

        def register_value(*, variable_name: str, value: object) -> None:
            found[variable_name] = value

        def change_installed_apps(
            *, variable_name: str, values: Sequence[object]
        ) -> Sequence[ast.expr]:
            value = self.info.installed_apps

            if "django.contrib.contenttypes" not in value:
                value = list(value)
                value.insert(0, "django.contrib.contenttypes")

            return [ast.Constant(value=app) for app in value if isinstance(app, str)]

        new_settings = file_changers.BasicPythonAssignmentChanger(
            cwd=options.cwd,
            root_dir=self.root_dir,
            path=settings_path,
            variable_changers={
                "SECRET_KEY": file_changers.VariableFinder(notify=register_value),
                "INSTALLED_APPS": file_changers.ListVariableChanger(change=change_installed_apps),
            },
        ).after_change(default_content="INSTALLED_APPS=[]")

        if "SECRET_KEY" not in found:
            new_settings = f"{new_settings}\nSECRET_KEY = '1'"

        monkeypatch_str = "import django_stubs_ext\ndjango_stubs_ext.monkeypatch()\n"
        new_settings = new_settings.replace(monkeypatch_str, "")
        if self.info.monkeypatch:
            new_settings = monkeypatch_str + new_settings

        if self.info.change_settings:
            new_settings = self.info.change_settings(new_settings)

        return new_settings

    def create_django_settings(
        self, *, options: protocols.RunOptions[Self], file_modification: protocols.FileModifier
    ) -> None:
        file_modification(
            path=self.info.mypy_configuration_filename,
            content=(
                self.info.mypy_configuration_content.format(
                    django_settings_module=self.info.django_settings_module
                )
                + "\n"
                + self.info.additional_mypy_configuration_content
            ),
        )

        settings_path = f"{self.info.django_settings_module}.py"
        settings_content = self.determine_django_settings_content(
            options=options, settings_path=settings_path
        )
        file_modification(
            path=settings_path,
            content=settings_content,
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class ScenarioRunner(scenarios.ScenarioRunner[Scenario]):
    def normalise_program_runner_notice(
        self, options: protocols.RunOptions[Scenario], notice: protocols.ProgramNotice, /
    ) -> protocols.ProgramNotice:
        if importlib.metadata.version("mypy") == "1.4.0":
            return notice.clone(
                msg=notice.msg.replace("Type[", "type[")
                .replace("django.db.models.query._QuerySet", "django.db.models.query.QuerySet")
                .replace("Type[Concrete?", "type[Concrete?")
            )
        else:
            return notice

    def execute_static_checking(
        self, options: protocols.RunOptions[Scenario]
    ) -> protocols.NoticeChecker[Scenario]:
        if self.scenario.info.debug:
            pathlib.Path("/tmp/debug").write_text("")
        else:
            pathlib.Path("/tmp/debug").unlink(missing_ok=True)

        self.scenario.create_django_settings(
            options=options, file_modification=self.file_modification
        )
        return super().execute_static_checking(options=options)


@dataclasses.dataclass(frozen=True, kw_only=True)
class ScenarioFile(builders.ScenarioFile):
    pass


@dataclasses.dataclass(frozen=True, kw_only=True)
class ScenarioBuilder(builders.ScenarioBuilder[Scenario, ScenarioFile]):
    def copy_django_apps(self, *apps: str) -> Self:
        def file_modification(path: str, content: str | None) -> None:
            self.on(path).set(content)

        for app in apps:
            self.scenario_runner.scenario.copy_django_app(app, file_modification)

        return self

    def set_installed_apps(self, *apps: str) -> Self:
        self.scenario_runner.scenario.info.installed_apps = list(apps)
        return self

    def set_and_copy_installed_apps(self, *apps: str) -> Self:
        self.set_installed_apps(*apps)
        self.copy_django_apps(*apps)
        return self

    def populate_virtual_deps(self, *, deps_dest: pathlib.Path) -> None:
        options = self.determine_options()
        self.scenario_runner.scenario.create_django_settings(
            options=options, file_modification=self.scenario_runner.file_modification
        )

        def populate(
            cwd: pathlib.Path, django_settings_module: str, deps_dest: pathlib.Path
        ) -> None:
            from extended_mypy_django_plugin.plugin import VirtualDependencyHandler

            class DependencyHandler(VirtualDependencyHandler):
                def interface_differentiator(self) -> str:
                    return "timestamp"

            DependencyHandler.create_report(
                project_root=cwd,
                django_settings_module=django_settings_module,
                virtual_deps_destination=deps_dest,
            )

        mainline = """
        if __name__ == "__main__":
            populate(pathlib.Path(sys.argv[1]), sys.argv[2], pathlib.Path(sys.argv[3]))
        """

        content = [
            "import pathlib",
            "import sys",
            textwrap.dedent(inspect.getsource(populate)),
            textwrap.dedent(mainline),
        ]

        with tempfile.TemporaryDirectory() as tmpdir:
            script = pathlib.Path(tmpdir, "script.py")
            script.write_text("\n".join(content))
            subprocess.check_call(
                [
                    sys.executable,
                    str(script),
                    str(options.cwd),
                    self.scenario_runner.scenario.info.django_settings_module,
                    str(deps_dest),
                ]
            )


if TYPE_CHECKING:
    _S: protocols.Scenario = cast(Scenario, None)
    _SF: protocols.P_ScenarioFile = cast(ScenarioFile, None)
    _SR: protocols.ScenarioRunner[Scenario] = cast(ScenarioRunner, None)



import pathlib
import subprocess
import sys
import tempfile
from collections.abc import MutableMapping
from itertools import chain
from typing import Generic

from mypy.options import Options
from mypy.plugin import Plugin as MypyPlugin

from . import plugin, protocols


class PluginProvider(Generic[protocols.T_Report]):
    """
    This can be used to provide both a mypy plugin as well as a __version__ that changes
    when mypy needs to do a full restart.

    Given either the extended_mypy_django_plugin.plugin.ExtendedMypyStubs class or a subclass
    of that, usage is::

        from extended_mypy_django_plugin.plugin import ExtendedMypyStubs, PluginProvider, VirtualDependencyHandler

        plugin = PluginProvider(ExtendedMypyStubs, VirtualDependencyHandler.create_report, locals())
    """

    previous_version: str

    def __init__(
        self,
        plugin_cls: type[plugin.ExtendedMypyStubs[protocols.T_Report]],
        virtual_dependency_handler: protocols.VirtualDependencyHandler[protocols.T_Report],
        locals: MutableMapping[str, object],
        /,
    ) -> None:
        self.locals = locals
        self.instance: plugin.ExtendedMypyStubs[protocols.T_Report] | None = None
        self.virtual_dependency_handler = virtual_dependency_handler
        self.plugin_cls = plugin_cls

    def __call__(self, version: str) -> type[MypyPlugin]:
        if self.instance is not None:
            # This only happens when using the mypy daemon
            # and also self.instance is only ever not None after previous_version is set
            # In this case if the options have changed then dmypy will already have restarted
            # So we can rely on the options on our previous instance
            self.set_new_version(
                self.determine_plugin_version(
                    options=self.instance.options, previous_version=self.previous_version
                )
            )

            # Inside dmypy, don't create a new plugin
            return MypyPlugin

        provider = self
        major, minor, _ = version.split(".", 2)

        def __init__(
            instance: plugin.ExtendedMypyStubs[protocols.T_Report], options: Options
        ) -> None:
            super(instance.__class__, instance).__init__(
                options,
                mypy_version_tuple=(int(major), int(minor)),  # type: ignore[call-arg]
                virtual_dependency_handler=self.virtual_dependency_handler,
            )
            provider.set_new_version(instance.virtual_dependency_report.version)
            provider.instance = instance

        return type("Plugin", (provider.plugin_cls,), {"__init__": __init__})

    def set_new_version(self, new_version: str) -> None:
        self.previous_version = new_version
        self.locals["__version__"] = new_version

    def determine_plugin_version(self, *, options: Options, previous_version: str) -> str:
        cmd = [
            sys.executable,
            "-m",
            "extended_mypy_django_plugin.scripts.determine_django_state",
            *(["--config-file", options.config_file] if options.config_file is not None else []),
            *chain.from_iterable(["--mypy-plugin", plugin] for plugin in options.plugins),
        ]

        err: subprocess.CalledProcessError | None = None

        with tempfile.NamedTemporaryFile() as fle:
            cmd.extend(["--version-file", fle.name])

            try:
                subprocess.run(cmd, capture_output=True, check=True)
            except subprocess.CalledProcessError as failed:
                err = failed
            else:
                return pathlib.Path(fle.name).read_text().strip()

        if err.returncode == 2:
            return previous_version
        else:
            message = [
                "",
                "Failed to determine information about the django setup",
                "",
                f"  > {' '.join(cmd)}",
                "  |",
            ]
            if err.stdout:
                for line in err.stdout.splitlines():
                    if isinstance(line, bytes):
                        line = line.decode()
                    message.append(f"  | {line}")
                if err.stderr:
                    message.append("  |")
            if err.stderr:
                for line in err.stderr.splitlines():
                    if isinstance(line, bytes):
                        line = line.decode()
                    message.append(f"  | {line}")
            message.append("  |")

            if previous_version:
                print("\n".join(message), file=sys.stderr)  # noqa: T201
                return previous_version
            else:
                raise RuntimeError("\n".join(message))



from extended_mypy_django_plugin.django_analysis import Project, discovery, protocols


class TestSettingsTypesDiscovery:
    def test_it_looks_at_values_to_determine_types(
        self, loaded_django_example: protocols.Loaded[Project]
    ) -> None:
        settings_types = discovery.NaiveSettingsTypesDiscovery[Project]()(loaded_django_example)
        assert settings_types == {
            "ABSOLUTE_URL_OVERRIDES": "<class 'dict'>",
            "ADMINS": "<class 'list'>",
            "ALLOWED_HOSTS": "<class 'list'>",
            "APPEND_SLASH": "<class 'bool'>",
            "AUTHENTICATION_BACKENDS": "<class 'list'>",
            "AUTH_PASSWORD_VALIDATORS": "<class 'list'>",
            "AUTH_USER_MODEL": "<class 'str'>",
            "BASE_DIR": "<class 'pathlib.PosixPath'>",
            "CACHES": "<class 'dict'>",
            "CACHE_MIDDLEWARE_ALIAS": "<class 'str'>",
            "CACHE_MIDDLEWARE_KEY_PREFIX": "<class 'str'>",
            "CACHE_MIDDLEWARE_SECONDS": "<class 'int'>",
            "CSRF_COOKIE_AGE": "<class 'int'>",
            "CSRF_COOKIE_DOMAIN": "<class 'NoneType'>",
            "CSRF_COOKIE_HTTPONLY": "<class 'bool'>",
            "CSRF_COOKIE_MASKED": "<class 'bool'>",
            "CSRF_COOKIE_NAME": "<class 'str'>",
            "CSRF_COOKIE_PATH": "<class 'str'>",
            "CSRF_COOKIE_SAMESITE": "<class 'str'>",
            "CSRF_COOKIE_SECURE": "<class 'bool'>",
            "CSRF_FAILURE_VIEW": "<class 'str'>",
            "CSRF_HEADER_NAME": "<class 'str'>",
            "CSRF_TRUSTED_ORIGINS": "<class 'list'>",
            "CSRF_USE_SESSIONS": "<class 'bool'>",
            "DATABASES": "<class 'dict'>",
            "DATABASE_ROUTERS": "<class 'list'>",
            "DATA_UPLOAD_MAX_MEMORY_SIZE": "<class 'int'>",
            "DATA_UPLOAD_MAX_NUMBER_FIELDS": "<class 'int'>",
            "DATA_UPLOAD_MAX_NUMBER_FILES": "<class 'int'>",
            "DATETIME_FORMAT": "<class 'str'>",
            "DATETIME_INPUT_FORMATS": "<class 'list'>",
            "DATE_FORMAT": "<class 'str'>",
            "DATE_INPUT_FORMATS": "<class 'list'>",
            "DEBUG": "<class 'bool'>",
            "DEBUG_PROPAGATE_EXCEPTIONS": "<class 'bool'>",
            "DECIMAL_SEPARATOR": "<class 'str'>",
            "DEFAULT_AUTO_FIELD": "<class 'str'>",
            "DEFAULT_CHARSET": "<class 'str'>",
            "DEFAULT_EXCEPTION_REPORTER": "<class 'str'>",
            "DEFAULT_EXCEPTION_REPORTER_FILTER": "<class 'str'>",
            "DEFAULT_FILE_STORAGE": "<class 'str'>",
            "DEFAULT_FROM_EMAIL": "<class 'str'>",
            "DEFAULT_INDEX_TABLESPACE": "<class 'str'>",
            "DEFAULT_TABLESPACE": "<class 'str'>",
            "DISALLOWED_USER_AGENTS": "<class 'list'>",
            "EMAIL_BACKEND": "<class 'str'>",
            "EMAIL_HOST": "<class 'str'>",
            "EMAIL_HOST_PASSWORD": "<class 'str'>",
            "EMAIL_HOST_USER": "<class 'str'>",
            "EMAIL_PORT": "<class 'int'>",
            "EMAIL_SSL_CERTFILE": "<class 'NoneType'>",
            "EMAIL_SSL_KEYFILE": "<class 'NoneType'>",
            "EMAIL_SUBJECT_PREFIX": "<class 'str'>",
            "EMAIL_TIMEOUT": "<class 'NoneType'>",
            "EMAIL_USE_LOCALTIME": "<class 'bool'>",
            "EMAIL_USE_SSL": "<class 'bool'>",
            "EMAIL_USE_TLS": "<class 'bool'>",
            "FILE_UPLOAD_DIRECTORY_PERMISSIONS": "<class 'NoneType'>",
            "FILE_UPLOAD_HANDLERS": "<class 'list'>",
            "FILE_UPLOAD_MAX_MEMORY_SIZE": "<class 'int'>",
            "FILE_UPLOAD_PERMISSIONS": "<class 'int'>",
            "FILE_UPLOAD_TEMP_DIR": "<class 'NoneType'>",
            "FIRST_DAY_OF_WEEK": "<class 'int'>",
            "FIXTURE_DIRS": "<class 'list'>",
            "FORCE_SCRIPT_NAME": "<class 'NoneType'>",
            "FORMAT_MODULE_PATH": "<class 'NoneType'>",
            "FORM_RENDERER": "<class 'str'>",
            "IGNORABLE_404_URLS": "<class 'list'>",
            "INSTALLED_APPS": "<class 'list'>",
            "INTERNAL_IPS": "<class 'list'>",
            "LANGUAGES": "<class 'list'>",
            "LANGUAGES_BIDI": "<class 'list'>",
            "LANGUAGE_CODE": "<class 'str'>",
            "LANGUAGE_COOKIE_AGE": "<class 'NoneType'>",
            "LANGUAGE_COOKIE_DOMAIN": "<class 'NoneType'>",
            "LANGUAGE_COOKIE_HTTPONLY": "<class 'bool'>",
            "LANGUAGE_COOKIE_NAME": "<class 'str'>",
            "LANGUAGE_COOKIE_PATH": "<class 'str'>",
            "LANGUAGE_COOKIE_SAMESITE": "<class 'NoneType'>",
            "LANGUAGE_COOKIE_SECURE": "<class 'bool'>",
            "LOCALE_PATHS": "<class 'list'>",
            "LOGGING": "<class 'dict'>",
            "LOGGING_CONFIG": "<class 'str'>",
            "LOGIN_REDIRECT_URL": "<class 'str'>",
            "LOGIN_URL": "<class 'str'>",
            "LOGOUT_REDIRECT_URL": "<class 'NoneType'>",
            "MANAGERS": "<class 'list'>",
            "MEDIA_ROOT": "<class 'str'>",
            "MEDIA_URL": "<class 'str'>",
            "MESSAGE_STORAGE": "<class 'str'>",
            "MIDDLEWARE": "<class 'list'>",
            "MIGRATION_MODULES": "<class 'dict'>",
            "MONTH_DAY_FORMAT": "<class 'str'>",
            "NUMBER_GROUPING": "<class 'int'>",
            "PASSWORD_HASHERS": "<class 'list'>",
            "PASSWORD_RESET_TIMEOUT": "<class 'int'>",
            "PREPEND_WWW": "<class 'bool'>",
            "ROOT_URLCONF": "<class 'str'>",
            "SECRET_KEY": "<class 'str'>",
            "SECRET_KEY_FALLBACKS": "<class 'list'>",
            "SECURE_CONTENT_TYPE_NOSNIFF": "<class 'bool'>",
            "SECURE_CROSS_ORIGIN_OPENER_POLICY": "<class 'str'>",
            "SECURE_HSTS_INCLUDE_SUBDOMAINS": "<class 'bool'>",
            "SECURE_HSTS_PRELOAD": "<class 'bool'>",
            "SECURE_HSTS_SECONDS": "<class 'int'>",
            "SECURE_PROXY_SSL_HEADER": "<class 'NoneType'>",
            "SECURE_REDIRECT_EXEMPT": "<class 'list'>",
            "SECURE_REFERRER_POLICY": "<class 'str'>",
            "SECURE_SSL_HOST": "<class 'NoneType'>",
            "SECURE_SSL_REDIRECT": "<class 'bool'>",
            "SERVER_EMAIL": "<class 'str'>",
            "SESSION_CACHE_ALIAS": "<class 'str'>",
            "SESSION_COOKIE_AGE": "<class 'int'>",
            "SESSION_COOKIE_DOMAIN": "<class 'NoneType'>",
            "SESSION_COOKIE_HTTPONLY": "<class 'bool'>",
            "SESSION_COOKIE_NAME": "<class 'str'>",
            "SESSION_COOKIE_PATH": "<class 'str'>",
            "SESSION_COOKIE_SAMESITE": "<class 'str'>",
            "SESSION_COOKIE_SECURE": "<class 'bool'>",
            "SESSION_ENGINE": "<class 'str'>",
            "SESSION_EXPIRE_AT_BROWSER_CLOSE": "<class 'bool'>",
            "SESSION_FILE_PATH": "<class 'NoneType'>",
            "SESSION_SAVE_EVERY_REQUEST": "<class 'bool'>",
            "SESSION_SERIALIZER": "<class 'str'>",
            "SETTINGS_MODULE": "<class 'str'>",
            "SHORT_DATETIME_FORMAT": "<class 'str'>",
            "SHORT_DATE_FORMAT": "<class 'str'>",
            "SIGNING_BACKEND": "<class 'str'>",
            "SILENCED_SYSTEM_CHECKS": "<class 'list'>",
            "STATICFILES_DIRS": "<class 'list'>",
            "STATICFILES_FINDERS": "<class 'list'>",
            "STATICFILES_STORAGE": "<class 'str'>",
            "STATIC_ROOT": "<class 'NoneType'>",
            "STATIC_URL": "<class 'str'>",
            "STORAGES": "<class 'dict'>",
            "TEMPLATES": "<class 'list'>",
            "TEST_NON_SERIALIZED_APPS": "<class 'list'>",
            "TEST_RUNNER": "<class 'str'>",
            "THOUSAND_SEPARATOR": "<class 'str'>",
            "TIME_FORMAT": "<class 'str'>",
            "TIME_INPUT_FORMATS": "<class 'list'>",
            "TIME_ZONE": "<class 'str'>",
            "UNIQUE_SETTING_TO_EXTENDED_MYPY_PLUGIN_DJANGOEXAMPLE": "<class 'str'>",
            "USE_DEPRECATED_PYTZ": "<class 'bool'>",
            "USE_I18N": "<class 'bool'>",
            "USE_L10N": "<class 'bool'>",
            "USE_THOUSAND_SEPARATOR": "<class 'bool'>",
            "USE_TZ": "<class 'bool'>",
            "USE_X_FORWARDED_HOST": "<class 'bool'>",
            "USE_X_FORWARDED_PORT": "<class 'bool'>",
            "WSGI_APPLICATION": "<class 'str'>",
            "X_FRAME_OPTIONS": "<class 'str'>",
            "YEAR_MONTH_FORMAT": "<class 'str'>",
        }



import functools
from typing import Generic, TypeVar

from mypy.nodes import Import, ImportAll, ImportFrom, MypyFile
from mypy.options import Options
from mypy.plugin import AnalyzeTypeContext, AttributeContext, MethodContext, ReportConfigContext
from mypy.types import Type as MypyType
from mypy_django_plugin import main
from mypy_django_plugin.transformers.managers import (
    resolve_manager_method,
    resolve_manager_method_from_instance,
)

from . import analyze, annotation_resolver, config, hook, protocols, type_checker

T_Report = TypeVar("T_Report", bound=protocols.Report)


class PlainHook(
    Generic[T_Report, hook.T_Ctx, hook.T_Ret],
    hook.PlainHook["ExtendedMypyStubs[T_Report]", hook.T_Ctx, hook.T_Ret],
):
    pass


class HookWithExtra(
    Generic[T_Report, hook.T_Ctx, hook.T_Extra, hook.T_Ret],
    hook.HookWithExtra["ExtendedMypyStubs[T_Report]", hook.T_Ctx, hook.T_Extra, hook.T_Ret],
):
    pass


class ExtendedMypyStubs(Generic[T_Report], main.NewSemanalDjangoPlugin):
    """
    The ``ExtendedMypyStubs`` mypy plugin extends the
    ``mypy_django_plugin.main.NewSemanalDjangoPlugin`` found in the active python
    environment.

    It implements the following mypy plugin hooks:

    .. automethod:: report_config_data

    .. automethod:: get_additional_deps

    .. autoattribute:: get_type_analyze_hook

    .. autoattribute:: get_attribute_hook

    .. autoattribute:: get_method_hook
    """

    @classmethod
    def make_virtual_dependency_report(
        cls,
        *,
        extra_options: config.ExtraOptions,
        virtual_dependency_handler: protocols.VirtualDependencyHandler[protocols.T_Report],
    ) -> protocols.CombinedReport[protocols.T_Report]:
        return virtual_dependency_handler(
            project_root=extra_options.project_root,
            django_settings_module=extra_options.django_settings_module,
            virtual_deps_destination=extra_options.scratch_path,
        )

    def __init__(
        self,
        options: Options,
        mypy_version_tuple: tuple[int, int],
        virtual_dependency_handler: protocols.VirtualDependencyHandler[T_Report],
    ) -> None:
        self.options = options
        self.extra_options = config.ExtraOptions.from_config(options.config_file)
        self.mypy_version_tuple = mypy_version_tuple

        self.virtual_dependency_report = self.make_virtual_dependency_report(
            extra_options=self.extra_options, virtual_dependency_handler=virtual_dependency_handler
        )

        make_resolver: protocols.ResolverMaker = functools.partial(
            annotation_resolver.make_resolver,
            get_concrete_aliases=self.virtual_dependency_report.report.get_concrete_aliases,
            get_queryset_aliases=self.virtual_dependency_report.report.get_queryset_aliases,
            plugin_lookup_fully_qualified=self.lookup_fully_qualified,
        )

        self.analyzer = analyze.Analyzer(make_resolver=make_resolver)
        self.type_checker = type_checker.TypeChecking(make_resolver=make_resolver)

        super().__init__(options)

        self.extra_init()

    def extra_init(self) -> None:
        """
        Place to add extra logic after __init__
        """

    def report_config_data(self, ctx: ReportConfigContext) -> dict[str, object]:
        """
        Add our extra options to the report config data, so that mypy knows to clear the cache
        if those settings change.
        """
        return {
            **super().report_config_data(ctx),
            "extended_mypy_django_plugin": self.extra_options.for_report(),
        }

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        """
        Ensure that models are re-analyzed if any other models that depend on
        them change.

        We use a generated "report" to re-analyze a file if a new dependency
        is discovered after this file has been processed.
        """
        file_import = file.fullname
        full_imports: set[str] = set()

        self.virtual_dependency_report.ensure_virtual_dependency(module_import_path=file.fullname)

        for imp in file.imports:
            if isinstance(imp, ImportFrom | ImportAll):
                if imp.relative:
                    prefix_base = ".".join(file_import.split(".")[: -imp.relative])
                    prefix = f"{prefix_base}.{imp.id}"
                else:
                    prefix = imp.id

                if isinstance(imp, ImportAll):
                    # This is the best we can do unfortunately
                    full_imports.add(prefix)
                else:
                    for name, _ in imp.names:
                        full_imports.add(f"{prefix}.{name}")

            elif isinstance(imp, Import):
                for name, _ in imp.ids:
                    full_imports.add(name)

        if self.options.use_fine_grained_cache:
            using_incremental_cache = False
        else:
            using_incremental_cache = (
                self.options.incremental and self.options.cache_dir != "/dev/null"
            )

        return list(
            self.virtual_dependency_report.report.additional_deps(
                file_import_path=file_import,
                imports=full_imports,
                django_settings_module=self.extra_options.django_settings_module,
                using_incremental_cache=using_incremental_cache,
                super_deps=super().get_additional_deps(file),
            )
        )

    @hook.hook
    class get_type_analyze_hook(
        HookWithExtra[T_Report, AnalyzeTypeContext, protocols.KnownAnnotations, MypyType]
    ):
        """
        Resolve classes annotated with ``Concrete`` or ``DefaultQuerySet``.
        """

        def choose(
            self, *, fullname: str, super_hook: hook.MypyHook[AnalyzeTypeContext, MypyType]
        ) -> hook.Choice[protocols.KnownAnnotations]:
            annotation = protocols.KnownAnnotations.resolve(fullname)
            if annotation is not None:
                return True, annotation
            else:
                return False

        def run(
            self,
            ctx: AnalyzeTypeContext,
            *,
            fullname: str,
            super_hook: hook.MypyHook[AnalyzeTypeContext, MypyType],
            extra: protocols.KnownAnnotations,
        ) -> MypyType:
            return self.plugin.analyzer.analyze_type(ctx, extra)

    @hook.hook
    class get_attribute_hook(PlainHook[T_Report, AttributeContext, MypyType]):
        """
        An implementation of the change found in
        https://github.com/typeddjango/django-stubs/pull/2027
        """

        def choose(
            self, *, fullname: str, super_hook: hook.MypyHook[AttributeContext, MypyType]
        ) -> bool:
            return super_hook is resolve_manager_method

        def run(self, ctx: AttributeContext) -> MypyType:
            return self.plugin.type_checker.extended_get_attribute_resolve_manager_method(
                ctx, resolve_manager_method_from_instance=resolve_manager_method_from_instance
            )

    @hook.hook
    class get_method_hook(PlainHook[T_Report, MethodContext, MypyType]):
        """
        Used to ensure Concrete.cast_as_concrete returns the appropriate type.
        """

        def choose(
            self, *, fullname: str, super_hook: hook.MypyHook[MethodContext, MypyType]
        ) -> bool:
            class_name, _, method_name = fullname.rpartition(".")
            if method_name == "cast_as_concrete":
                info = self.plugin._get_typeinfo_or_none(class_name)
                if info and info.has_base(protocols.KnownClasses.CONCRETE.value):
                    return True

            return False

        def run(self, ctx: MethodContext) -> MypyType:
            return self.plugin.type_checker.modify_cast_as_concrete(ctx)



from .concrete_models import ConcreteModelsDiscovery
from .container import Discovery
from .import_path import ImportPath, InvalidImportPath
from .known_models import DefaultInstalledModulesDiscovery, make_module_creator
from .settings_types import NaiveSettingsTypesDiscovery

__all__ = [
    "Discovery",
    "ImportPath",
    "InvalidImportPath",
    "make_module_creator",
    "DefaultInstalledModulesDiscovery",
    "ConcreteModelsDiscovery",
    "NaiveSettingsTypesDiscovery",
]



import os
import pathlib

from extended_mypy_django_plugin.django_analysis import hasher

tests_dir = pathlib.Path(__file__).parent.parent


class TestAdler32Hash:
    def test_it_creates_a_consistent_hash(self) -> None:
        found: dict[str, str] = {}
        contents: set[bytes] = set()
        for root, _, files in os.walk(tests_dir):
            for name in files:
                made: set[str] = set()
                content = (pathlib.Path(root) / name).read_bytes()
                if content in contents:
                    continue
                contents.add(content)

                for i in range(10):
                    made.add(hasher.adler32_hash(*content.splitlines()))

                # naive check to show the hash is consistent
                assert len(made) == 1
                hsh = next(iter(made))
                assert len(hsh) < 32
                found[name] = hsh

        # naive check to show the hash is different for different content
        assert len(set(found)) > 5
        assert len(set(found)) == len(set(found.values()))



from . import discovery, protocols, virtual_dependencies
from .discovery.import_path import ImportPath
from .fields import Field
from .hasher import adler32_hash
from .models import Model
from .modules import Module
from .project import Discovered, Loaded, Project, replaced_env_vars_and_sys_path

__all__ = [
    "protocols",
    "discovery",
    "virtual_dependencies",
    "Field",
    "ImportPath",
    "adler32_hash",
    "Model",
    "Module",
    "Project",
    "Loaded",
    "Discovered",
    "replaced_env_vars_and_sys_path",
]



"""
The helpers under ``extended_mypy_django_plugin.plugin.hooks`` exist as a helper
for defining hooks on a mypy plugin.

The way a mypy plugin works is there is a class that inherits from
``mypy.plugin.Plugin`` with specific hooks that take in a string and returns
a function.

So for example, the ``get_customize_class_mro_hook`` hook will take in the
``fullname`` representing classes that can be altered, and a function that takes
in a ``ClassDefContext`` object and returns ``None`` must be returned if
the plugin wants to do something with that object.

The first plugin that mypy encounters which returns a function will win and no
other plugins will get to look at that fullname.

This is fine but it can get awkward when the function returned takes in more
options than only the context. To improve this situation, this module implements
a ``Hook`` class and associated decorator to turn those hooks into python
descriptors that do the correct thing.

.. code-block:: python

    from typing import Generic
    from mypy.plugin import Plugin, AttributeContext
    from mypy.types import Type as MypyType
    from extended_mypy_django_plugin.plugin import hook


    class MyPlugin(Plugin):
        @hook.hook
        class get_attribute_hook(hook.Hook["MyPlugin", AttributeContext, MypyType]):
            def choose(
                self, *, fullname: str, super_hook: hook.MypyHook[AttributeContext, MypyType]
            ) -> bool:
                # return True if we want to use the run method for ``self.fullname``.
                # With super_hook being the result of `super().get_attribute_hook(fullname)` on the plugin
                return self.fullname.endswith(".blah")

            def run(self, ctx: AttributeContext) -> MypyType:
                # Do stuff
                return ...

The :class:`extended_mypy_django_plugin.plugin.hook.hook` decorator turns that
attribute into a descriptor that will cache an instance
of the ``Hook`` with the instance of the plugin and that hook on the parent class
(useful when the mypy plugin is subclassing an existing mypy plugin)

The descriptor will then return the ``hook`` method on the Hook, which matches the required
signature for the hook.

There are two default implementations of this Hook class:

.. autoclass:: PlainHook

.. autoclass:: HookWithExtra

And they must be used with the ``hook`` decorator:

.. autoclass:: hook(hook_kls: type[Hook[T_Plugin, T_Ctx, T_Ret]])
"""

from __future__ import annotations

import abc
import dataclasses
import functools
from collections.abc import Callable
from typing import Generic, Literal, TypeVar, overload

from mypy.plugin import Plugin

T_Ctx = TypeVar("T_Ctx")
T_Ret = TypeVar("T_Ret")
T_Extra = TypeVar("T_Extra")
T_Plugin = TypeVar("T_Plugin", bound=Plugin)


Choice = Literal[False] | tuple[Literal[True], T_Extra]
MypyHook = Callable[[T_Ctx], T_Ret] | None
MypyHookMaker = Callable[[str], MypyHook[T_Ctx, T_Ret]]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Hook(Generic[T_Plugin, T_Ctx, T_Ret], abc.ABC):
    """
    Class used to represent both the choosing and running logic for a hook
    on a mypy Plugin.

    This is to be used with the ``hook`` descriptor defined below. See the
    docstring on the module for more information.

    Concrete subclasses of this must implement ``hook``.
    """

    plugin: T_Plugin
    super_hook_maker: MypyHookMaker[T_Ctx, T_Ret]

    @abc.abstractmethod
    def hook(self, fullname: str) -> MypyHook[T_Ctx, T_Ret]:
        """
        This is the function that mypy ends up calling when asking the plugin
        if it should handle something.
        """


@dataclasses.dataclass(frozen=True, kw_only=True)
class HookWithExtra(
    Generic[T_Plugin, T_Ctx, T_Extra, T_Ret], Hook[T_Plugin, T_Ctx, T_Ret], abc.ABC
):
    """
    This is an implementation of Hook that allows passing information from the choose
    method into the run method.

    Mypy plugins work by implementing methods that take in a string and either return a callable
    that takes action, or returns None. Whichever plugin returns a function first wins.

    This hook class splits the two parts into two methods: deciding if we should choose this hook
    and the running of the logic.

    This implementation in particular also allows passing information from the choosing part into the running
    part:

    .. code-block:: python

        from typing import Generic
        from mypy.plugin import Plugin, AttributeContext
        from mypy.types import Type as MypyType
        from extended_mypy_django_plugin.plugin import hook


        class MyPlugin(Plugin):
            @hook.hook
            class get_attribute_hook(hook.HookWithExtra["MyPlugin", AttributeContext, str, MypyType]):
                def choose(
                    self, *, fullname: str, super_hook: hook.MypyHook[AttributeContext, MypyType]
                ) -> hook.Choice[str]:
                    if fullname.startswith("things"):
                        return True, "one"
                    elif fullname.startswith("other"):
                        return True, "two"
                    else:
                        return False

                def run(
                    self,
                    ctx: AttributeContext,
                    *,
                    fullname: str,
                    super_hook: hook.MypyHook[AttributeContext, MypyType],
                    extra: str,
                ) -> MypyType:
                    # at this point, extra will either be "one" or "two"

                    # Do stuff
                    return ...

    The type of the data passed between ``choose`` and ``run`` is specified when filling out the Generic.

    When the plugin runs the ``hook`` method on this implementation, it will determine what
    the parent class would return from this hook, get a result from ``choose`` from the ``fullname``
    and that ``super_hook`` and either return ``super_hook`` if the ``choose`` returned ``False``
    or return ``functools.partial(self.run, fullname=fullname, super_hook=super_hook, extra=extra)``
    where extra is the second item in a ``(True, extra)`` tuple returned from ``choose``.
    """

    @abc.abstractmethod
    def run(
        self,
        ctx: T_Ctx,
        *,
        fullname: str,
        super_hook: MypyHook[T_Ctx, T_Ret],
        extra: T_Extra,
    ) -> T_Ret: ...

    @abc.abstractmethod
    def choose(self, *, fullname: str, super_hook: MypyHook[T_Ctx, T_Ret]) -> Choice[T_Extra]: ...

    def hook(self, fullname: str) -> MypyHook[T_Ctx, T_Ret]:
        """
        This is the function that mypy ends up calling when asking the plugin
        if it should handle something.
        """
        super_hook = self.super_hook_maker(fullname)
        result = self.choose(fullname=fullname, super_hook=super_hook)
        if result is False:
            return super_hook
        else:
            return functools.partial(
                self.run, fullname=fullname, super_hook=super_hook, extra=result[1]
            )


@dataclasses.dataclass(frozen=True, kw_only=True)
class PlainHook(Hook[T_Plugin, T_Ctx, T_Ret], abc.ABC):
    """
    This is an implementation of ``Hook`` that does the bare minimum.

    It's ``hook`` method will determine what would have been returned for this hook from
    the parent class and pass the ``fullname`` and that ``super_hook`` to ``choose``.

    If ``choose`` returns ``False``, then ``hook`` returns the ``super_hook``, otherwise
    it returns ``self.run``.
    """

    @abc.abstractmethod
    def run(self, ctx: T_Ctx) -> T_Ret: ...

    @abc.abstractmethod
    def choose(self, *, fullname: str, super_hook: MypyHook[T_Ctx, T_Ret]) -> bool: ...

    def hook(self, fullname: str) -> MypyHook[T_Ctx, T_Ret]:
        """
        This is the function that mypy ends up calling when asking the plugin
        if it should handle something.
        """
        super_hook = self.super_hook_maker(fullname)
        result = self.choose(fullname=fullname, super_hook=super_hook)
        if result is False:
            return super_hook
        else:
            return self.run


@dataclasses.dataclass
class hook(Generic[T_Plugin, T_Ctx, T_Ret]):
    """
    This is a class decorator used to return a callable object that takes in
    a string and either returns a function that mypy can use to perform an action,
    or return None if this hook does not need to do anything in that instance.

    It is used as a decorator for a subclass of ``Hook`` and will cache an instance
    of that hook between multiple access of the descriptor
    """

    hook_kls: type[Hook[T_Plugin, T_Ctx, T_Ret]]

    name: str = dataclasses.field(init=False)
    owner: type = dataclasses.field(init=False)

    _hook_instance: Hook[T_Plugin, T_Ctx, T_Ret] | None = dataclasses.field(
        default=None, init=False
    )

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name
        self.owner = owner
        self.__doc__ = self.hook_kls.__doc__

    @overload
    def __get__(self, instance: None, owner: None) -> hook[T_Plugin, T_Ctx, T_Ret]: ...

    @overload
    def __get__(
        self, instance: T_Plugin, owner: type[T_Plugin]
    ) -> MypyHookMaker[T_Ctx, T_Ret]: ...

    def __get__(
        self, instance: T_Plugin | None, owner: type[T_Plugin] | None = None
    ) -> hook[T_Plugin, T_Ctx, T_Ret] | MypyHookMaker[T_Ctx, T_Ret]:
        if instance is None:
            return self

        if self._hook_instance is None:
            self._hook_instance = self.hook_kls(
                # We use self.owner cause we want the class where the descriptor was used
                # Rather than the class of the instance the attribute was accessed on
                plugin=instance,
                super_hook_maker=getattr(super(self.owner, instance), self.name),
            )

        return self._hook_instance.hook


__all__ = ["Hook", "PlainHook", "HookWithExtra", "hook", "Choice", "MypyHook", "MypyHookMaker"]



import pathlib
import textwrap

import pytest

from extended_mypy_django_plugin.django_analysis import ImportPath
from extended_mypy_django_plugin.plugin import ExtraOptions


class TestGetExtraOptions:
    def test_it_can_get_options_from_ini_file(self, tmp_path: pathlib.Path) -> None:
        config = tmp_path / "config.ini"

        config.write_text(
            textwrap.dedent("""
        [mypy.plugins.django-stubs]
        project_root = $MYPY_CONFIG_FILE_DIR/project
        scratch_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main
        django_settings_module = my.settings
        """)
        )

        expected_scratch = tmp_path / ".mypy_django_scratch" / "main"
        assert not expected_scratch.exists()

        assert ExtraOptions.from_config(config) == ExtraOptions(
            project_root=tmp_path / "project",
            django_settings_module=ImportPath("my.settings"),
            scratch_path=expected_scratch,
        )

        assert expected_scratch.exists()

    def test_it_can_get_options_from_toml_file(self, tmp_path: pathlib.Path) -> None:
        config = tmp_path / "pyproject.toml"

        config.write_text(
            textwrap.dedent("""
        [tool.django-stubs]
        project_root = "$MYPY_CONFIG_FILE_DIR/project"
        scratch_path = "$MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main"
        django_settings_module = "my.settings"
        """)
        )

        expected_scratch = tmp_path / ".mypy_django_scratch" / "main"
        assert not expected_scratch.exists()

        assert ExtraOptions.from_config(config) == ExtraOptions(
            project_root=tmp_path / "project",
            scratch_path=expected_scratch,
            django_settings_module=ImportPath("my.settings"),
        )

        assert expected_scratch.exists()

    def test_complains_if_django_settings_module_is_not_specified(
        self, tmp_path: pathlib.Path
    ) -> None:
        versions = (
            (
                "mypy.ini",
                """
                [mypy.plugins.django-stubs]
                scratch_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main
                """,
            ),
            (
                "pyproject.toml",
                """
                [tool.django-stubs]
                scratch_path = "$MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main"
                """,
            ),
        )

        for name, content in versions:
            config = tmp_path / name
            config.write_text(textwrap.dedent(content))

            with pytest.raises(
                ValueError,
                match="Please specify 'django_settings_module' in the django-stubs section of your mypy configuration",
            ):
                ExtraOptions.from_config(config)

    def test_complains_if_django_settings_module_is_not_valid(
        self, tmp_path: pathlib.Path
    ) -> None:
        versions = (
            (
                "mypy.ini",
                """
                [mypy.plugins.django-stubs]
                scratch_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main
                django_settings_module = -asdf
                """,
            ),
            (
                "pyproject.toml",
                """
                [tool.django-stubs]
                scratch_path = "$MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main"
                django_settings_module = "-asdf"
                """,
            ),
        )

        for name, content in versions:
            config = tmp_path / name
            config.write_text(textwrap.dedent(content))

            with pytest.raises(
                ValueError,
                match="Provided path was not a valid python import path: '-asdf'",
            ):
                ExtraOptions.from_config(config)

    def test_complains_if_scratch_path_not_specified(self, tmp_path: pathlib.Path) -> None:
        versions = (
            (
                "mypy.ini",
                """
                [mypy.plugins.django-stubs]
                django_settings_module = my.settings
                """,
            ),
            (
                "pyproject.toml",
                """
                [tool.django-stubs]
                django_settings_module = "my.settings"
                """,
            ),
        )

        for name, content in versions:
            config = tmp_path / name
            config.write_text(textwrap.dedent(content))

            with pytest.raises(
                ValueError,
                match="Please specify 'scratch_path' in the django-stubs section of your mypy configuration",
            ):
                ExtraOptions.from_config(config)

    def test_complains_if_config_file_is_none(self) -> None:
        with pytest.raises(SystemExit):
            ExtraOptions.from_config(None)

    def test_complains_if_config_file_is_invalid_format(self, tmp_path: pathlib.Path) -> None:
        versions = (
            (
                "mypy.ini",
                """
                [mypy.plugins.django-stubs
                scratch_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main
                django_settings_module = my.settings
                """,
            ),
            (
                "pyproject.toml",
                """
                [tool.django-stubs
                scratch_path = "$MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/main"
                django_settings_module = "my.settings"
                """,
            ),
        )

        for name, content in versions:
            config = tmp_path / name
            config.write_text(textwrap.dedent(content))

            with pytest.raises(SystemExit):
                ExtraOptions.from_config(config)

    def test_complains_if_config_file_is_missing_django_stubs_section(
        self, tmp_path: pathlib.Path
    ) -> None:
        versions = (
            (
                "mypy.ini",
                """
                [mypy.plugins.not-correct]
                hello = there
                django_settings_module = my.settings
                """,
            ),
            (
                "pyproject.toml",
                """
                [tool.not-correct]
                hello = "there"
                django_settings_module = "my.settings"
                """,
            ),
        )

        for name, content in versions:
            config = tmp_path / name
            config.write_text(textwrap.dedent(content))

            with pytest.raises(SystemExit):
                ExtraOptions.from_config(config)



import dataclasses
import pathlib
import re
import sys
import textwrap

import pytest
from extended_mypy_django_plugin_test_driver import (
    Scenario,
    ScenarioBuilder,
    ScenarioRunner,
    assertions,
)
from pytest_typing_runner import expectations, protocols, runners


@dataclasses.dataclass(frozen=True, kw_only=True)
class SimpleExpectation:
    options: protocols.RunOptions[Scenario]

    @dataclasses.dataclass(frozen=True, kw_only=True)
    class Expectation:
        result: protocols.RunResult
        notice_checker: protocols.NoticeChecker[Scenario]

        def check(self) -> None:
            raise NotImplementedError()

    expectation: type[Expectation]

    @classmethod
    def setup(cls, expectation: type[Expectation], /) -> protocols.ExpectationsSetup[Scenario]:
        def make(
            *, options: protocols.RunOptions[Scenario]
        ) -> protocols.ExpectationsMaker[Scenario]:
            return cls(expectation=expectation, options=options)

        return make

    def __call__(self) -> protocols.Expectations[Scenario]:
        Expectation = self.expectation

        class Expected:
            def check(self, *, notice_checker: protocols.NoticeChecker[Scenario]) -> None:
                Expectation(notice_checker=notice_checker, result=notice_checker.result).check()

        return Expected()


class TestErrors:
    def test_it_complains_if_annotating_a_typevar(self, builder: ScenarioBuilder) -> None:
        @builder.run_and_check_after
        def _() -> None:
            builder.set_installed_apps("example")
            builder.on("example/__init__.py").set("")

            builder.on("example/apps.py").set(
                """
                from django.apps import AppConfig

                class Config(AppConfig):
                    name = "example"
                """,
            )

            builder.on("example/models.py").set(
                """
                from __future__ import annotations

                from django.db import models
                from typing import TypeVar
                from typing_extensions import Self
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                T_Leader = TypeVar("T_Leader", bound="Concrete[Leader]")

                class Leader(models.Model):
                    @classmethod
                    def new(cls) -> Concrete[Self]:
                        # ^ ERROR(misc) ^ Using a concrete annotation on a TypeVar is not currently supported
                        raise NotImplementedError()

                    class Meta:
                        abstract = True

                class Follower1(Leader):
                    pass

                def make_leader(model: type[T_Leader]) -> Concrete[T_Leader]:
                    # ^ ERROR(misc) ^ Using a concrete annotation on a TypeVar is not currently supported
                    raise NotImplementedError()

                def make_qs(model: type[T_Leader]) -> DefaultQuerySet[T_Leader]:
                    # ^ ERROR(misc) ^ Using a concrete annotation on a TypeVar is not currently supported
                    raise NotImplementedError()
                """,
            )

    def test_gracefully_handles_determine_version_failure_on_startup(
        self,
        scenario: Scenario,
        scenario_runner: ScenarioRunner,
        tmp_path: pathlib.Path,
    ) -> None:
        options = runners.RunOptions.create(scenario_runner)
        if not options.program_runner_maker.is_daemon:
            pytest.skip("Test only relevant for the daemon")

        plugin_provider = tmp_path / "plugin.py"

        plugin_provider.write_text(
            textwrap.dedent("""
            import pathlib

            from extended_mypy_django_plugin.django_analysis import Project
            from extended_mypy_django_plugin.plugin import PluginProvider, VirtualDependencyHandler, ExtendedMypyStubs


            class VirtualDependencyHandler(VirtualDependencyHandler):
                @classmethod
                def make_project(
                    cls, *, project_root: pathlib.Path, django_settings_module: str
                ) -> Project:
                    raise ValueError("Computer says no")


            plugin = PluginProvider(ExtendedMypyStubs, VirtualDependencyHandler.create_report, locals())
            """)
        )

        scenario.info.mypy_configuration_content = textwrap.dedent(
            f"""
            [mypy]
            plugins = {plugin_provider}
            mypy_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/test

            [mypy.plugins.django-stubs]
            django_settings_module = mysettings
            scratch_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/test
            """
        )

        @dataclasses.dataclass(frozen=True, kw_only=True)
        class CheckFails(SimpleExpectation.Expectation):
            def check(self) -> None:
                assert self.result.exit_code != 0

                assertions.assert_glob_lines(
                    self.result.stdout + self.result.stderr,
                    f"""
                Error constructing plugin instance of Plugin
                
                Daemon crashed!
                Traceback (most recent call last):
                File "*extended_mypy_django_plugin/_plugin/plugin.py", line *, in make_virtual_dependency_report
                File "{plugin_provider}", line *, in make_project
                ValueError: Computer says no
                """,
                )

        scenario_runner.run_and_check(
            setup_expectations=SimpleExpectation.setup(CheckFails), options=options
        )

    def test_gracefully_handles_determine_version_failure_on_subsequent_run(
        self, scenario: Scenario, scenario_runner: ScenarioRunner, tmp_path: pathlib.Path
    ) -> None:
        options = runners.RunOptions.create(scenario_runner)
        if not options.program_runner_maker.is_daemon:
            pytest.skip("Test only relevant for the daemon")

        plugin_provider = tmp_path / "plugin.py"
        marker = tmp_path / "marker"
        marker2 = tmp_path / "marker2"

        # pytest plugin I use needs work which is under way but in the meantime I must hack around
        # how inside the test I can't turn off the auto second try
        marker.write_text("")
        marker2.write_text("")

        # Changing the contents of this file will trigger the daemon to restart
        # So we instead rely on the existence or absence of a file to trigger the error
        plugin_provider.write_text(
            textwrap.dedent(f"""
            import pathlib

            from extended_mypy_django_plugin.django_analysis import Project
            from extended_mypy_django_plugin import main


            class VirtualDependencyHandler(main.VirtualDependencyHandler):
                @classmethod
                def make_project(
                    cls, *, project_root: pathlib.Path, django_settings_module: str
                ) -> Project:
                    if pathlib.Path("{marker}").exists():
                        pathlib.Path("{marker}").unlink()
                        return super().make_project(
                            project_root=project_root,
                            django_settings_module=django_settings_module,
                        )

                    if pathlib.Path("{marker2}").exists():
                        pathlib.Path("{marker2}").unlink()
                        return super().make_project(
                            project_root=project_root,
                            django_settings_module=django_settings_module,
                        )

                    # Make this only fail on the startup to show if the run after restart works then
                    # then this failing doesn't break the daemon
                    pathlib.Path("{marker}").write_text('')
                    raise ValueError("Computer says no")


            plugin = main.PluginProvider(main.ExtendedMypyStubs, VirtualDependencyHandler.create_report, locals())
        """)
        )

        scenario.info.mypy_configuration_content = textwrap.dedent(
            f"""
            [mypy]
            plugins = {plugin_provider}
            mypy_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/test

            [mypy.plugins.django-stubs]
            django_settings_module = mysettings
            scratch_path = $MYPY_CONFIG_FILE_DIR/.mypy_django_scratch/test
            """
        )

        scenario_runner.run_and_check(
            setup_expectations=expectations.Expectations.setup_for_success, options=options
        )

        called: list[int] = []

        @dataclasses.dataclass(frozen=True, kw_only=True)
        class CheckNoCrashShowsFailure(SimpleExpectation.Expectation):
            def check(self) -> None:
                called.append(self.result.exit_code)

                assert self.result.exit_code == 0
                command = (
                    f"{sys.executable} -m extended_mypy_django_plugin.scripts.determine_django_state"
                    f" --config-file mypy.ini"
                    f" --mypy-plugin {plugin_provider}"
                    " --version-file *"
                )

                assertions.assert_glob_lines(
                    self.result.stdout + self.result.stderr,
                    f"""
                    Failed to determine information about the django setup

                    > {command}
                    |
                    | Traceback (most recent call last):
                    |   File "{plugin_provider}", line *, in make_project
                    |     raise ValueError("Computer says no")
                    | ValueError: Computer says no
                    |
                    """,
                )

        scenario_runner.run_and_check(
            setup_expectations=SimpleExpectation.setup(CheckNoCrashShowsFailure), options=options
        )
        assert called == [0]

        @dataclasses.dataclass(frozen=True, kw_only=True)
        class CheckNoOutput(SimpleExpectation.Expectation):
            def check(self) -> None:
                called.append(self.result.exit_code)

                assert self.result.exit_code == 0
                assert re.match(
                    "Success: no issues found in \d+ source files?",
                    (self.result.stdout + self.result.stderr).strip(),
                )

        marker.write_text("")
        scenario_runner.run_and_check(
            setup_expectations=SimpleExpectation.setup(CheckNoOutput), options=options
        )
        assert called == [0, 0]

    def test_knowing_types_of_fields_on_parent_classes(
        self, scenario: Scenario, builder: ScenarioBuilder
    ) -> None:
        """
        This is a regression test to ensure that get_additional_deps doesn't cause class
        definitions to not understand parent types
        """

        @builder.run_and_check_after
        def _() -> None:
            builder.set_installed_apps("example", "example2")
            for app in ("example", "example2"):
                builder.on(f"{app}/__init__.py").set("")
                builder.on(f"{app}/apps.py").set(
                    f"""
                    from django.apps import AppConfig

                    class Config(AppConfig):
                        name = "{app}"
                    """,
                )

            builder.on("example/models/__init__.py").set(
                """
                from .parent import Parent 
                """
            )
            builder.on("example2/models/__init__.py").set(
                """
                from .children import Child
                """
            )

            builder.on("example/models/parent.py").set(
                """
                from django.db import models

                class Parent(models.Model):
                    response_body = models.TextField(max_length=12, blank=True)

                    class Meta:
                        abstract = True
                """,
            )

            builder.on("example2/models/children.py").set(
                """
                from example import models as common_models
                from typing import TYPE_CHECKING
                from django.db import models

                class Child(common_models.Parent):
                    response_body = models.BooleanField()
                    # ^ ERROR(assignment) ^ Incompatible types in assignment (expression has type "BooleanField[bool | Combinable, bool]", base class "Parent" defined the type as "TextField[str | Combinable, str]")
                """,
            )



import dataclasses
import functools
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Generic, TypedDict, cast

from typing_extensions import Self

from .. import project, protocols


@dataclasses.dataclass
class VirtualDependencySummary:
    virtual_namespace: protocols.ImportPath
    virtual_import_path: protocols.ImportPath
    module_import_path: protocols.ImportPath
    significant_info: Sequence[str]


@dataclasses.dataclass
class VirtualDependency(Generic[protocols.T_Project]):
    module: protocols.Module
    summary: VirtualDependencySummary
    all_related_models: Sequence[protocols.ImportPath]
    concrete_models: protocols.ConcreteModelsMap

    @classmethod
    def create(
        cls,
        *,
        discovered_project: protocols.Discovered[protocols.T_Project],
        module: protocols.Module,
        virtual_dependency_namer: protocols.VirtualDependencyNamer,
    ) -> Self:
        concrete_models = {
            import_path: discovered_project.concrete_models[import_path]
            for import_path in module.defined_models
        }

        related_models: set[protocols.ImportPath] = set()
        for model in module.defined_models.values():
            related_models.add(model.import_path)
            for field in model.all_fields.values():
                if field.related_model:
                    related_models.add(field.related_model)

        return cls(
            module=module,
            summary=VirtualDependencySummary(
                virtual_namespace=virtual_dependency_namer.namespace,
                virtual_import_path=virtual_dependency_namer(module.import_path),
                module_import_path=module.import_path,
                significant_info=list(
                    cls.find_significant_info_from_module(
                        discovered_project=discovered_project,
                        module=module,
                        concrete_models=concrete_models,
                    )
                ),
            ),
            all_related_models=sorted(related_models),
            concrete_models=concrete_models,
        )

    @classmethod
    def find_significant_info_from_module(
        cls,
        *,
        discovered_project: protocols.Discovered[protocols.T_Project],
        module: protocols.Module,
        concrete_models: protocols.ConcreteModelsMap,
    ) -> Iterator[str]:
        prefix = f"module:{module.import_path}"
        yield prefix
        for model_import_path, concrete_children in concrete_models.items():
            yield f"{prefix}>concrete:{model_import_path}={','.join(conc.import_path for conc in concrete_children)}"

        for model in module.defined_models.values():
            for info in cls.find_significant_info_from_model(
                discovered_project=discovered_project, module=module, model=model
            ):
                yield f"{prefix}>{info}"

    @classmethod
    def find_significant_info_from_model(
        cls,
        *,
        discovered_project: protocols.Discovered[protocols.T_Project],
        module: protocols.Module,
        model: protocols.Model,
    ) -> Iterator[str]:
        model_prefix = f"model:{model.import_path}"
        yield f"{model_prefix}>is_abstract:{model.is_abstract}"

        if model.default_custom_queryset:
            yield f"{model_prefix}>custom_queryset:{model.default_custom_queryset}"

        for i, mro_import_path in enumerate(model.models_in_mro):
            yield f"{model_prefix}>mro_{i}:{mro_import_path}"

        for name, field in model.all_fields.items():
            field_prefix = f"{model_prefix}>field:{name}"
            yield field_prefix
            for info in cls.find_significant_info_from_field(
                discovered_project=discovered_project, module=module, model=model, field=field
            ):
                yield f"{field_prefix}>{info}"

    @classmethod
    def find_significant_info_from_field(
        cls,
        *,
        discovered_project: protocols.Discovered[protocols.T_Project],
        module: protocols.Module,
        model: protocols.Model,
        field: protocols.Field,
    ) -> Iterator[str]:
        yield f"field_type:{field.field_type}"
        if field.related_model:
            yield f"related_model:{field.related_model}"


if TYPE_CHECKING:
    C_VirtualDependency = VirtualDependency[project.C_Project]

    _VD: protocols.VirtualDependency = cast(VirtualDependency[protocols.P_Project], None)
    _VDS: protocols.VirtualDependencySummary = cast(VirtualDependencySummary, None)

    class _RequiredMakerKwargs(TypedDict):
        virtual_dependency_namer: protocols.VirtualDependencyNamer

    _VDM: protocols.P_VirtualDependencyMaker = functools.partial(
        VirtualDependency[protocols.P_Project].create, **cast(_RequiredMakerKwargs, None)
    )

    _CVDM: protocols.VirtualDependencyMaker[project.C_Project, C_VirtualDependency] = (
        functools.partial(
            VirtualDependency[project.C_Project].create, **cast(_RequiredMakerKwargs, None)
        )
    )



import configparser
import dataclasses
import pathlib
import sys
from collections.abc import Mapping

from mypy_django_plugin import config as django_stubs_config
from typing_extensions import Self

from ..django_analysis import ImportPath, protocols
from ..version import VERSION

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


@dataclasses.dataclass(frozen=True)
class ExtraOptions:
    """
    The extended_mypy_django_plugin adds two options to the django-stubs configuration in the mypy configuration

    scratch_path
        A folder where virtual dependencies are written to

    project_root
        This defaults to the folder the config is found in. It should be the path to the root of
        the project. This value will be added to sys.path before Django is loaded

    django_settings_module
        The option used to set DJANGO_SETTINGS_MODULE when loading django
    """

    scratch_path: pathlib.Path
    project_root: pathlib.Path
    django_settings_module: protocols.ImportPath

    @classmethod
    def from_config(cls, filepath: str | pathlib.Path | None) -> Self:
        if filepath is None:
            django_stubs_config.exit_with_error(django_stubs_config.INVALID_FILE)

        config_path = pathlib.Path(filepath)
        return cls.from_options(options=_parse_mypy_config(config_path), filepath=config_path)

    @classmethod
    def from_options(cls, *, options: Mapping[str, object], filepath: pathlib.Path) -> Self:
        """
        Construct the extra options from the mypy configuration
        """
        scratch_path = _sanitize_path(filepath, options, "scratch_path", required=True)
        assert scratch_path is not None

        project_root = _sanitize_path(filepath, options, "project_root")
        if project_root is None:
            project_root = filepath.parent

        django_settings_module_value = _sanitize_str(
            filepath, options, "django_settings_module", required=True
        )
        assert django_settings_module_value is not None
        django_settings_module = ImportPath(django_settings_module_value)

        scratch_path.mkdir(parents=True, exist_ok=True)

        return cls(
            scratch_path=scratch_path,
            project_root=project_root,
            django_settings_module=django_settings_module,
        )

    def for_report(self) -> dict[str, str]:
        """
        Get the options that were found to be used for the mypy report_config_data hook
        """
        return {
            "scratch_path": str(self.scratch_path),
            "project_root": str(self.project_root),
            "django_settings_module": self.django_settings_module,
            "plugin_version": str(VERSION),
        }


def _parse_mypy_config(filepath: pathlib.Path) -> Mapping[str, object]:
    if filepath.suffix == ".toml":
        return _parse_toml_config(filepath)
    else:
        return _parse_ini_config(filepath)


def _parse_toml_config(filepath: pathlib.Path) -> Mapping[str, object]:
    try:
        with filepath.open(mode="rb") as f:
            data = tomllib.load(f)
    except (tomllib.TOMLDecodeError, OSError):
        django_stubs_config.exit_with_error(django_stubs_config.COULD_NOT_LOAD_FILE, is_toml=True)

    if not isinstance(tool := data.get("tool"), Mapping) or not isinstance(
        result := tool.get("django-stubs"), Mapping
    ):
        django_stubs_config.exit_with_error(
            django_stubs_config.MISSING_SECTION.format(section="tool.django-stubs"),
            is_toml=True,
        )

    return result


def _parse_ini_config(filepath: pathlib.Path) -> Mapping[str, object]:
    parser = configparser.ConfigParser()
    try:
        with filepath.open(encoding="utf-8") as f:
            parser.read_file(f, source=str(filepath))
    except (configparser.ParsingError, OSError):
        django_stubs_config.exit_with_error(django_stubs_config.COULD_NOT_LOAD_FILE)

    section = "mypy.plugins.django-stubs"
    if not parser.has_section(section):
        django_stubs_config.exit_with_error(
            django_stubs_config.MISSING_SECTION.format(section=section)
        )

    return dict(parser.items("mypy.plugins.django-stubs"))


def _sanitize_str(
    config_path: pathlib.Path,
    options: Mapping[str, object],
    option: str,
    *,
    required: bool = False,
) -> str | None:
    if not isinstance(value := options.get(option), str):
        if required:
            raise ValueError(
                f"Please specify '{option}' in the django-stubs section of your mypy configuration ({config_path})"
            )
        else:
            return None

    while value and value.startswith('"'):
        value = value[1:]
    while value and value.endswith('"'):
        value = value[:-1]

    return value


def _sanitize_path(
    config_path: pathlib.Path,
    options: Mapping[str, object],
    option: str,
    *,
    required: bool = False,
) -> pathlib.Path | None:
    value = _sanitize_str(config_path, options, option, required=required)
    if value is None:
        return None
    return pathlib.Path(value.replace("$MYPY_CONFIG_FILE_DIR", str(config_path.parent)))



def interface____differentiated__2() -> None:
    return None

mod = "django.contrib.auth.models"
summary = "__virtual__.mod_2289830437::django.contrib.auth.models::installed_apps=__installed_apps_hash__::significant=1922836542"

import django.contrib.auth.models
import django.db.models
ConcreteQuerySet__AbstractUser = django.db.models.QuerySet[django.contrib.auth.models.User]
ConcreteQuerySet__Group = django.db.models.QuerySet[django.contrib.auth.models.Group]
ConcreteQuerySet__Permission = django.db.models.QuerySet[django.contrib.auth.models.Permission]
ConcreteQuerySet__PermissionsMixin = django.db.models.QuerySet[django.contrib.auth.models.User]
ConcreteQuerySet__User = django.db.models.QuerySet[django.contrib.auth.models.User]
Concrete__AbstractUser = django.contrib.auth.models.User
Concrete__Group = django.contrib.auth.models.Group
Concrete__Permission = django.contrib.auth.models.Permission
Concrete__PermissionsMixin = django.contrib.auth.models.User
Concrete__User = django.contrib.auth.models.User



import dataclasses

import pytest

from extended_mypy_django_plugin.django_analysis import (
    Field,
    ImportPath,
    Model,
    protocols,
    virtual_dependencies,
)


class TestCombiningReports:
    def test_it_can_combine_reports(self) -> None:
        report1 = virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("P1"): ImportPath("CP1"),
                ImportPath("C1"): ImportPath("CC1"),
            },
            concrete_querysets={
                ImportPath("P1"): ImportPath("CP1QS"),
                ImportPath("C1"): ImportPath("CC1QS"),
            },
            report_import_path={
                ImportPath("M1"): ImportPath("VM1"),
            },
        )
        report2 = virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("C2"): ImportPath("CC2"),
            },
            concrete_querysets={
                ImportPath("C2"): ImportPath("C2QS"),
            },
            report_import_path={
                ImportPath("M2"): ImportPath("VM2"),
            },
        )
        report3 = virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("P2"): ImportPath("CP2"),
                ImportPath("C3"): ImportPath("CC3"),
                ImportPath("C4"): ImportPath("CC4"),
            },
            concrete_querysets={
                ImportPath("P2"): ImportPath("CP2QS"),
                ImportPath("C3"): ImportPath("CC3QS"),
                ImportPath("C4"): ImportPath("CC4QS"),
            },
            report_import_path={
                ImportPath("M3"): ImportPath("VM3"),
            },
        )

        def write_empty_virtual_dep(
            *, module_import_path: protocols.ImportPath
        ) -> protocols.ImportPath | None:
            raise NotImplementedError()

        final_report = virtual_dependencies.ReportCombiner(
            report_maker=virtual_dependencies.Report, reports=(report1, report2, report3)
        ).combine(version="__version__", write_empty_virtual_dep=write_empty_virtual_dep)

        assert final_report.version == "__version__"
        assert final_report.report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("P1"): ImportPath("CP1"),
                ImportPath("C1"): ImportPath("CC1"),
                ImportPath("C2"): ImportPath("CC2"),
                ImportPath("P2"): ImportPath("CP2"),
                ImportPath("C3"): ImportPath("CC3"),
                ImportPath("C4"): ImportPath("CC4"),
            },
            concrete_querysets={
                ImportPath("P1"): ImportPath("CP1QS"),
                ImportPath("C1"): ImportPath("CC1QS"),
                ImportPath("C2"): ImportPath("C2QS"),
                ImportPath("P2"): ImportPath("CP2QS"),
                ImportPath("C3"): ImportPath("CC3QS"),
                ImportPath("C4"): ImportPath("CC4QS"),
            },
            report_import_path={
                ImportPath("M1"): ImportPath("VM1"),
                ImportPath("M2"): ImportPath("VM2"),
                ImportPath("M3"): ImportPath("VM3"),
            },
        )

    def test_it_can_ensure_empty_vritual_deps(self) -> None:
        written: list[tuple[protocols.ImportPath, str | None]] = []
        report1 = virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("P1"): ImportPath("CP1"),
                ImportPath("C1"): ImportPath("CC1"),
            },
            concrete_querysets={
                ImportPath("P1"): ImportPath("CP1QS"),
                ImportPath("C1"): ImportPath("CC1QS"),
            },
            report_import_path={
                ImportPath("M1"): ImportPath("VM1"),
            },
        )
        report2 = virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("C2"): ImportPath("CC2"),
            },
            concrete_querysets={
                ImportPath("C2"): ImportPath("C2QS"),
            },
            report_import_path={
                ImportPath("M2"): ImportPath("VM2"),
            },
        )

        def write_empty_virtual_dep(
            *, module_import_path: protocols.ImportPath
        ) -> protocols.ImportPath | None:
            if module_import_path == ImportPath("E1"):
                # This doesn't happen cause ".models." not in the path
                raise RuntimeError("Shouldn't get this far")
            elif module_import_path == ImportPath("E1.models"):
                written.append((module_import_path, "end models"))
                return ImportPath("VE1.models")
            elif module_import_path == ImportPath("E1.models.things"):
                written.append((module_import_path, "inside models"))
                return ImportPath("VE1.models.things")
            else:
                written.append((module_import_path, None))
                return None

        final_report = virtual_dependencies.ReportCombiner(
            report_maker=virtual_dependencies.Report, reports=(report1, report2)
        ).combine(version="__version__", write_empty_virtual_dep=write_empty_virtual_dep)

        assert final_report.version == "__version__"
        expected = virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("P1"): ImportPath("CP1"),
                ImportPath("C1"): ImportPath("CC1"),
                ImportPath("C2"): ImportPath("CC2"),
            },
            concrete_querysets={
                ImportPath("P1"): ImportPath("CP1QS"),
                ImportPath("C1"): ImportPath("CC1QS"),
                ImportPath("C2"): ImportPath("C2QS"),
            },
            report_import_path={
                ImportPath("M1"): ImportPath("VM1"),
                ImportPath("M2"): ImportPath("VM2"),
            },
        )
        assert final_report.report == expected

        assert written == []

        # Can get content, but doesn't pass our naming heuristic
        final_report.ensure_virtual_dependency(module_import_path=ImportPath("E1"))
        assert written == []
        assert final_report.report == expected

        # passes naming heuristic of ending in .models
        final_report.ensure_virtual_dependency(module_import_path=(e2 := ImportPath("E1.models")))
        assert written == [(e2, "end models")]
        assert final_report.report != expected
        expected.report_import_path[e2] = ImportPath("VE1.models")
        assert final_report.report == expected

        # passes naming heuristic of containing .models.
        final_report.ensure_virtual_dependency(
            module_import_path=(e3 := ImportPath("E1.models.things"))
        )
        assert written == [(e2, "end models"), (e3, "inside models")]
        assert final_report.report != expected
        expected.report_import_path[e3] = ImportPath("VE1.models.things")
        assert final_report.report == expected

        # Passes heuristics but doesn't return a name
        final_report.ensure_virtual_dependency(
            module_import_path=(e4 := ImportPath("E2.models.ignoreme"))
        )
        assert written == [(e2, "end models"), (e3, "inside models"), (e4, None)]
        assert final_report.report == expected


class TestBuildingReport:
    def test_registering_module_edits_report_import_path(self) -> None:
        report = virtual_dependencies.Report()
        report.register_module(
            module_import_path=ImportPath("one.two"),
            virtual_import_path=ImportPath("virtual.one.two"),
        )
        assert report == virtual_dependencies.Report(
            report_import_path={
                ImportPath("one.two"): ImportPath("virtual.one.two"),
            }
        )

        report.register_module(
            module_import_path=ImportPath("three.four"),
            virtual_import_path=ImportPath("virtual.three.four"),
        )
        assert report == virtual_dependencies.Report(
            report_import_path={
                ImportPath("one.two"): ImportPath("virtual.one.two"),
                ImportPath("three.four"): ImportPath("virtual.three.four"),
            }
        )

    @dataclasses.dataclass
    class BuildingScenario:
        parent: protocols.Model = dataclasses.field(
            default_factory=lambda: Model(
                model_name="Parent",
                module_import_path=ImportPath("my.parents"),
                import_path=ImportPath("my.parents.Parent"),
                is_abstract=True,
                default_custom_queryset=None,
                all_fields={},
                models_in_mro=[],
            )
        )

        def register_parent(self, report: virtual_dependencies.Report) -> None:
            report.register_model(
                model_import_path=self.parent.import_path,
                virtual_import_path=ImportPath(f"virtual.{self.parent.module_import_path}"),
                concrete_name="Concrete__Parent",
                concrete_queryset_name="QuerySet__Parent",
                concrete_models=[self.model1, self.model2],
            )

        model1: protocols.Model = dataclasses.field(
            default_factory=lambda: Model(
                model_name="Model1",
                module_import_path=ImportPath("my.models"),
                import_path=ImportPath("my.models.Model1"),
                is_abstract=False,
                default_custom_queryset=ImportPath("my.querysets.Model1QS"),
                all_fields={},
                models_in_mro=[ImportPath("my.parents.Parent")],
            )
        )

        def register_model1(self, report: virtual_dependencies.Report) -> None:
            report.register_model(
                model_import_path=self.model1.import_path,
                virtual_import_path=ImportPath(f"virtual.{self.model1.module_import_path}"),
                concrete_name="Concrete__Model1",
                concrete_queryset_name="QuerySet__Model1",
                concrete_models=[self.model1],
            )

        model2: protocols.Model = dataclasses.field(
            default_factory=lambda: Model(
                model_name="Model2",
                module_import_path=ImportPath("my.models"),
                import_path=ImportPath("my.models.Model2"),
                is_abstract=False,
                default_custom_queryset=None,
                all_fields={
                    "one": Field(
                        model_import_path=ImportPath("my.models.Model2"),
                        field_type=ImportPath("fields.Foreign"),
                        related_model=ImportPath("my.models.Model1"),
                    ),
                    "two": Field(
                        model_import_path=ImportPath("my.models.Model2"),
                        field_type=ImportPath("fields.Foreign"),
                        related_model=ImportPath("other.models.Model3"),
                    ),
                },
                models_in_mro=[ImportPath("my.parents.Parent")],
            )
        )

        def register_model2(self, report: virtual_dependencies.Report) -> None:
            report.register_model(
                model_import_path=self.model2.import_path,
                virtual_import_path=ImportPath(f"virtual.{self.model2.module_import_path}"),
                concrete_name="Concrete__Model2",
                concrete_queryset_name="QuerySet__Model2",
                concrete_models=[self.model2],
            )

        model3: protocols.Model = dataclasses.field(
            default_factory=lambda: Model(
                model_name="Model3",
                module_import_path=ImportPath("other.models"),
                import_path=ImportPath("other.models.Model3"),
                is_abstract=False,
                default_custom_queryset=ImportPath("my.querysets.Model3QS"),
                all_fields={
                    "three": Field(
                        model_import_path=ImportPath("other.models.Model3"),
                        field_type=ImportPath("fields.Foreign"),
                        related_model=ImportPath("more.models.Model4"),
                    ),
                },
                models_in_mro=[ImportPath("mixins.BlahMixin")],
            )
        )

        def register_model3(self, report: virtual_dependencies.Report) -> None:
            report.register_model(
                model_import_path=self.model3.import_path,
                virtual_import_path=ImportPath(f"virtual.{self.model3.module_import_path}"),
                concrete_name="Concrete__Model3",
                concrete_queryset_name="QuerySet__Model3",
                concrete_models=[self.model3],
            )

    def test_building_individually_parent(self) -> None:
        scenario = self.BuildingScenario()

        report = virtual_dependencies.Report()
        scenario.register_parent(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.Concrete__Parent"),
            },
            concrete_querysets={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.QuerySet__Parent"),
            },
        )

    def test_building_individually_model1(self) -> None:
        scenario = self.BuildingScenario()

        report = virtual_dependencies.Report()
        scenario.register_model1(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.Concrete__Model1"),
            },
            concrete_querysets={
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.QuerySet__Model1"),
            },
        )

    def test_building_individually_model2(self) -> None:
        scenario = self.BuildingScenario()

        report = virtual_dependencies.Report()
        scenario.register_model2(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("my.models.Model2"): ImportPath("virtual.my.models.Concrete__Model2"),
            },
            concrete_querysets={
                ImportPath("my.models.Model2"): ImportPath("virtual.my.models.QuerySet__Model2"),
            },
        )

    def test_building_individually_model3(self) -> None:
        scenario = self.BuildingScenario()

        report = virtual_dependencies.Report()
        scenario.register_model3(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("other.models.Model3"): ImportPath(
                    "virtual.other.models.Concrete__Model3"
                ),
            },
            concrete_querysets={
                ImportPath("other.models.Model3"): ImportPath(
                    "virtual.other.models.QuerySet__Model3"
                ),
            },
        )

    def test_building_up_a_report(self) -> None:
        scenario = self.BuildingScenario()

        report = virtual_dependencies.Report()
        scenario.register_parent(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.Concrete__Parent"),
            },
            concrete_querysets={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.QuerySet__Parent"),
            },
        )

        scenario.register_model1(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.Concrete__Parent"),
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.Concrete__Model1"),
            },
            concrete_querysets={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.QuerySet__Parent"),
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.QuerySet__Model1"),
            },
        )

        scenario.register_model2(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.Concrete__Parent"),
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.Concrete__Model1"),
                ImportPath("my.models.Model2"): ImportPath("virtual.my.models.Concrete__Model2"),
            },
            concrete_querysets={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.QuerySet__Parent"),
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.QuerySet__Model1"),
                ImportPath("my.models.Model2"): ImportPath("virtual.my.models.QuerySet__Model2"),
            },
        )

        scenario.register_model3(report)

        assert report == virtual_dependencies.Report(
            concrete_annotations={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.Concrete__Parent"),
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.Concrete__Model1"),
                ImportPath("my.models.Model2"): ImportPath("virtual.my.models.Concrete__Model2"),
                ImportPath("other.models.Model3"): ImportPath(
                    "virtual.other.models.Concrete__Model3"
                ),
            },
            concrete_querysets={
                ImportPath("my.parents.Parent"): ImportPath("virtual.my.parents.QuerySet__Parent"),
                ImportPath("my.models.Model1"): ImportPath("virtual.my.models.QuerySet__Model1"),
                ImportPath("my.models.Model2"): ImportPath("virtual.my.models.QuerySet__Model2"),
                ImportPath("other.models.Model3"): ImportPath(
                    "virtual.other.models.QuerySet__Model3"
                ),
            },
        )

    @pytest.mark.parametrize("using_incremental_cache", (True, False))
    def test_additional_deps(self, using_incremental_cache: bool) -> None:
        report = virtual_dependencies.Report(
            report_import_path={
                ImportPath(k): ImportPath(v)
                for k, v in {
                    "one.two": "v_one_two",
                    "three.four": "v_three_four",
                    "five.six": "v_five_six",
                    "six.seven": "v_six_seven",
                    "eight.nine": "v_eight_nine",
                    "twelve.thirteen": "v_twelve_thirteen",
                    "another.one": "v_another_one",
                    "more": "v_more",
                }.items()
            }
        )

        ##
        ## These are the test cases above but they still only return super_deps

        # File name startswith django., it is effectively ignored
        made = report.additional_deps(
            file_import_path="django.db.models",
            imports=set(),
            super_deps=(super_deps := [(25, "one.two", -1), (25, "two", -1)]),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        assert sorted(made) == sorted(super_deps)

        # Expansion depending only on super deps and imports
        made = report.additional_deps(
            file_import_path="some.place",
            imports=set(),
            super_deps=(super_deps := []),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        assert sorted(made) == sorted(super_deps)

        made = report.additional_deps(
            file_import_path="some.place",
            imports={"eight.nine", "typing.Protocol"},
            super_deps=(super_deps := []),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        assert sorted(made) == sorted(super_deps)

        made = report.additional_deps(
            file_import_path="some.place",
            imports=set(),
            super_deps=(super_deps := [(25, "eight.nine", -1), (25, "typing.Protocol", 13)]),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        assert sorted(made) == sorted(super_deps)

        made = report.additional_deps(
            file_import_path="some.place",
            imports={"one.two"},
            super_deps=(super_deps := [(25, "hello.there", -1), (25, "typing.Protocol", 13)]),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        assert sorted(made) == sorted(super_deps)

        # Also add from the file import itself
        made = report.additional_deps(
            file_import_path="another.one",
            imports={"one.two"},
            super_deps=(super_deps := [(25, "hello.there", -1), (25, "typing.Protocol", 13)]),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        if using_incremental_cache:
            assert sorted(made) == sorted(
                [*super_deps, (25, "v_another_one", -1), (25, "my.settings", -1)]
            )
        else:
            assert sorted(made) == sorted([*super_deps, (25, "v_another_one", -1)])

        made = report.additional_deps(
            file_import_path="another.one",
            imports={"one.two.MyModel"},
            super_deps=(super_deps := [(25, "hello.there", -1), (25, "typing.Protocol", 13)]),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        if using_incremental_cache:
            assert sorted(made) == sorted(
                [*super_deps, (25, "v_another_one", -1), (25, "my.settings", -1)]
            )
        else:
            assert sorted(made) == sorted([*super_deps, (25, "v_another_one", -1)])

        # Also virtual_deps themselves don't add extra
        made = report.additional_deps(
            file_import_path="v_another_one",
            imports={"one.two.MyModel"},
            super_deps=(super_deps := [(25, "hello.there", -1), (25, "typing.Protocol", 13)]),
            django_settings_module="my.settings",
            using_incremental_cache=using_incremental_cache,
        )
        assert sorted(made) == sorted(super_deps)



import dataclasses
from typing import TYPE_CHECKING, cast

from .. import protocols
from ..discovery.import_path import ImportPath


@dataclasses.dataclass
class VirtualDependencyNamer:
    namespace: protocols.ImportPath
    hasher: protocols.Hasher

    def __call__(self, module: protocols.ImportPath, /) -> protocols.ImportPath:
        return ImportPath(f"{self.namespace}.mod_{self.hasher(module.encode())}")


if TYPE_CHECKING:
    _VDN: protocols.VirtualDependencyNamer = cast(VirtualDependencyNamer, None)



import functools
import importlib
import pathlib
import shutil
import textwrap

import pytest

from extended_mypy_django_plugin.django_analysis import (
    ImportPath,
    Project,
    adler32_hash,
    protocols,
    virtual_dependencies,
)

ReportSummaryGetter = virtual_dependencies.ReportSummaryGetter


class TestVirtualDependencyScribe:
    class TestGetSummary:
        @pytest.fixture
        def get_report_summary(self) -> ReportSummaryGetter:
            return virtual_dependencies.VirtualDependencyScribe.get_report_summary

        def test_it_says_no_to_directories(
            self, tmp_path: pathlib.Path, get_report_summary: ReportSummaryGetter
        ) -> None:
            assert get_report_summary(tmp_path) is None

        def test_it_says_no_if_the_path_doesnt_end_in_python_extension(
            self, tmp_path: pathlib.Path, get_report_summary: ReportSummaryGetter
        ) -> None:
            (without_extension := tmp_path / "one").write_text(
                'mod = "extended_mypy_django_plugin.version"\nsummary = "stuff"'
            )

            assert get_report_summary(without_extension) is None

            with_extension = tmp_path / "one.py"
            shutil.move(without_extension, with_extension)

            assert get_report_summary(with_extension) == "stuff"

        def test_it_says_yes_to_files_and_links(
            self, tmp_path: pathlib.Path, get_report_summary: ReportSummaryGetter
        ) -> None:
            (one := tmp_path / "one.py").write_text(
                'mod = "extended_mypy_django_plugin.version"\nsummary = "trees"'
            )
            (two := tmp_path / "two.py").symlink_to(one)

            assert get_report_summary(one) == "trees"
            assert get_report_summary(two) == "trees"

        def test_it_says_no_if_cant_find_mod(
            self, tmp_path: pathlib.Path, get_report_summary: ReportSummaryGetter
        ) -> None:
            (location := tmp_path / "one.py").write_text('summary = "things"')
            assert get_report_summary(location) is None

            with open(location, "a") as fle:
                fle.write('\nmod = "extended_mypy_django_plugin.version"')

            assert get_report_summary(location) == "things"

        def test_it_says_no_if_cant_find_summary(
            self, tmp_path: pathlib.Path, get_report_summary: ReportSummaryGetter
        ) -> None:
            (location := tmp_path / "one.py").write_text(
                'mod = "extended_mypy_django_plugin.version"'
            )
            assert get_report_summary(location) is None

            with open(location, "a") as fle:
                fle.write('\nsummary = "blah"')

            assert get_report_summary(location) == "blah"

        def test_it_says_no_if_cant_import_mod(
            self, tmp_path: pathlib.Path, get_report_summary: ReportSummaryGetter
        ) -> None:
            with pytest.raises(ModuleNotFoundError):
                importlib.import_module("does.not.exist")

            (location := tmp_path / "one.py").write_text(
                'mod = "does.not.exist\nsummary = "lalalala"'
            )
            assert get_report_summary(location) is None

            location.write_text('mod = "pytest"\nsummary = "ladelala"')
            assert get_report_summary(location) == "ladelala"

    class TestWrite:
        class Scenario:
            def __init__(self, discovered_project: protocols.Discovered[Project]) -> None:
                self.count: int = 0

                virtual_dependency_maker = functools.partial(
                    virtual_dependencies.VirtualDependency[Project].create,
                    virtual_dependency_namer=virtual_dependencies.VirtualDependencyNamer(
                        namespace=ImportPath("__virtual__"), hasher=adler32_hash
                    ),
                )

                self.all_virtual_dependencies = virtual_dependencies.VirtualDependencyGenerator(
                    virtual_dependency_maker=virtual_dependency_maker
                )(discovered_project=discovered_project)

            def scribe(
                self,
                *,
                hasher: protocols.Hasher,
                virtual_dependency: virtual_dependencies.VirtualDependency[Project],
            ) -> virtual_dependencies.RenderedVirtualDependency[virtual_dependencies.Report]:
                def make_differentiator() -> str:
                    self.count += 1
                    return f"__differentiated__{self.count}"

                return virtual_dependencies.VirtualDependencyScribe(
                    hasher=hasher,
                    report_maker=virtual_dependencies.Report,
                    virtual_dependency=virtual_dependency,
                    all_virtual_dependencies=self.all_virtual_dependencies,
                    installed_apps_hash="__installed_apps_hash__",
                    make_differentiator=make_differentiator,
                ).render()

            def make_report(
                self,
                *,
                concrete_annotations: dict[str, str],
                concrete_querysets: dict[str, str],
                report_import_path: dict[str, str],
            ) -> virtual_dependencies.Report:
                """
                Helper to make the tests below easier to read
                """
                return virtual_dependencies.Report(
                    concrete_annotations={
                        ImportPath(k): ImportPath(v) for k, v in concrete_annotations.items()
                    },
                    concrete_querysets={
                        ImportPath(k): ImportPath(v) for k, v in concrete_querysets.items()
                    },
                    report_import_path={
                        ImportPath(k): ImportPath(v) for k, v in report_import_path.items()
                    },
                )

        def test_writing_virtual_dependencies_1(
            self, discovered_django_example: protocols.Discovered[Project]
        ) -> None:
            scenario = self.Scenario(discovered_django_example)

            hasher_called: list[int] = []

            def hasher(*parts: bytes) -> str:
                hasher_called.append(1)
                assert parts == (
                    b"module:djangoexample.exampleapp2.models",
                    b"module:djangoexample.exampleapp2.models>concrete:djangoexample.exampleapp2.models.ChildOther=djangoexample.exampleapp2.models.ChildOther",
                    b"module:djangoexample.exampleapp2.models>concrete:djangoexample.exampleapp2.models.ChildOther2=djangoexample.exampleapp2.models.ChildOther2",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>is_abstract:False",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>mro_0:djangoexample.exampleapp.models.Parent",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>field:id",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>field:id>field_type:django.db.models.fields.BigAutoField",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>field:one",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>field:one>field_type:django.db.models.fields.CharField",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>field:two",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther>field:two>field_type:django.db.models.fields.CharField",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>is_abstract:False",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>mro_0:djangoexample.exampleapp.models.Parent",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>field:id",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>field:id>field_type:django.db.models.fields.BigAutoField",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>field:one",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>field:one>field_type:django.db.models.fields.CharField",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>field:two",
                    b"module:djangoexample.exampleapp2.models>model:djangoexample.exampleapp2.models.ChildOther2>field:two>field_type:django.db.models.fields.CharField",
                )
                return "__hashed_for_great_good__"

            virtual_dependency = scenario.all_virtual_dependencies[
                ImportPath("djangoexample.exampleapp2.models")
            ]

            content = textwrap.dedent("""
            def interface____differentiated__1() -> None:
                return None

            mod = "djangoexample.exampleapp2.models"
            summary = "__virtual__.mod_3537308831::djangoexample.exampleapp2.models::installed_apps=__installed_apps_hash__::significant=__hashed_for_great_good__"

            import django.db.models
            import djangoexample.exampleapp2.models
            ConcreteQuerySet__ChildOther = django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther]
            ConcreteQuerySet__ChildOther2 = django.db.models.QuerySet[djangoexample.exampleapp2.models.ChildOther2]
            Concrete__ChildOther = djangoexample.exampleapp2.models.ChildOther
            Concrete__ChildOther2 = djangoexample.exampleapp2.models.ChildOther2
            """).strip()

            summary_hash = (
                "__virtual__.mod_3537308831"
                "::djangoexample.exampleapp2.models"
                "::installed_apps=__installed_apps_hash__"
                "::significant=__hashed_for_great_good__"
            )

            written = scenario.scribe(hasher=hasher, virtual_dependency=virtual_dependency)
            assert written == virtual_dependencies.RenderedVirtualDependency(
                content=content + "\n",
                summary_hash=summary_hash,
                report=scenario.make_report(
                    concrete_annotations={
                        "djangoexample.exampleapp2.models.ChildOther": "__virtual__.mod_3537308831.Concrete__ChildOther",
                        "djangoexample.exampleapp2.models.ChildOther2": "__virtual__.mod_3537308831.Concrete__ChildOther2",
                    },
                    concrete_querysets={
                        "djangoexample.exampleapp2.models.ChildOther": "__virtual__.mod_3537308831.ConcreteQuerySet__ChildOther",
                        "djangoexample.exampleapp2.models.ChildOther2": "__virtual__.mod_3537308831.ConcreteQuerySet__ChildOther2",
                    },
                    report_import_path={
                        "djangoexample.exampleapp2.models": "__virtual__.mod_3537308831"
                    },
                ),
                virtual_import_path=virtual_dependency.summary.virtual_import_path,
            )

            assert hasher_called == [1]

        def test_writing_virtual_dependencies_2(
            self, discovered_django_example: protocols.Discovered[Project]
        ) -> None:
            scenario = self.Scenario(discovered_django_example)

            hasher_called: list[int] = []

            def hasher(*parts: bytes) -> str:
                hasher_called.append(1)
                assert parts == (
                    b"module:djangoexample.relations1.models",
                    b"module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Abstract=djangoexample.relations1.models.Child1,djangoexample.relations1.models.Child2",
                    b"module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Child1=djangoexample.relations1.models.Child1",
                    b"module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Child2=djangoexample.relations1.models.Child2",
                    b"module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Concrete1=djangoexample.relations1.models.Concrete1",
                    b"module:djangoexample.relations1.models>concrete:djangoexample.relations1.models.Concrete2=djangoexample.relations1.models.Concrete2",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Abstract>is_abstract:True",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>is_abstract:False",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>custom_queryset:djangoexample.relations1.models.Child1QuerySet",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>mro_0:djangoexample.relations1.models.Abstract",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:Concrete2_children+",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:Concrete2_children+>field_type:django.db.models.fields.reverse_related.ManyToOneRel",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:children",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:children>field_type:django.db.models.fields.reverse_related.ManyToManyRel",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:children>related_model:djangoexample.relations1.models.Concrete2",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:id",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child1>field:id>field_type:django.db.models.fields.BigAutoField",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>is_abstract:False",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>mro_0:djangoexample.relations1.models.Abstract",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>field:id",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Child2>field:id>field_type:django.db.models.fields.BigAutoField",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>is_abstract:False",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>custom_queryset:djangoexample.relations1.models.Concrete1QuerySet",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:c2s",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:c2s>field_type:django.db.models.fields.reverse_related.ManyToOneRel",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:c2s>related_model:djangoexample.relations1.models.Concrete2",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:thing",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:thing>field_type:django.db.models.fields.reverse_related.OneToOneRel",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:thing>related_model:djangoexample.relations2.models.Thing",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:id",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete1>field:id>field_type:django.db.models.fields.BigAutoField",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>is_abstract:False",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:Concrete2_children+",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:Concrete2_children+>field_type:django.db.models.fields.reverse_related.ManyToOneRel",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:id",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:id>field_type:django.db.models.fields.BigAutoField",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:concrete1",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:concrete1>field_type:django.db.models.fields.related.ForeignKey",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:concrete1>related_model:djangoexample.relations1.models.Concrete1",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:children",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:children>field_type:django.db.models.fields.related.ManyToManyField",
                    b"module:djangoexample.relations1.models>model:djangoexample.relations1.models.Concrete2>field:children>related_model:djangoexample.relations1.models.Child1",
                )
                return "__hashed_for_greater_good__"

            virtual_dependency = scenario.all_virtual_dependencies[
                ImportPath("djangoexample.relations1.models")
            ]

            content = textwrap.dedent("""
            def interface____differentiated__1() -> None:
                return None

            mod = "djangoexample.relations1.models"
            summary = "__virtual__.mod_3327724610::djangoexample.relations1.models::installed_apps=__installed_apps_hash__::significant=__hashed_for_greater_good__"

            import django.db.models
            import djangoexample.relations1.models
            ConcreteQuerySet__Abstract = djangoexample.relations1.models.Child1QuerySet | django.db.models.QuerySet[djangoexample.relations1.models.Child2]
            ConcreteQuerySet__Child1 = djangoexample.relations1.models.Child1QuerySet
            ConcreteQuerySet__Child2 = django.db.models.QuerySet[djangoexample.relations1.models.Child2]
            ConcreteQuerySet__Concrete1 = djangoexample.relations1.models.Concrete1QuerySet
            ConcreteQuerySet__Concrete2 = django.db.models.QuerySet[djangoexample.relations1.models.Concrete2]
            Concrete__Abstract = djangoexample.relations1.models.Child1 | djangoexample.relations1.models.Child2
            Concrete__Child1 = djangoexample.relations1.models.Child1
            Concrete__Child2 = djangoexample.relations1.models.Child2
            Concrete__Concrete1 = djangoexample.relations1.models.Concrete1
            Concrete__Concrete2 = djangoexample.relations1.models.Concrete2
            """).strip()

            summary_hash = (
                "__virtual__.mod_3327724610"
                "::djangoexample.relations1.models"
                "::installed_apps=__installed_apps_hash__"
                "::significant=__hashed_for_greater_good__"
            )

            written = scenario.scribe(hasher=hasher, virtual_dependency=virtual_dependency)
            assert written == virtual_dependencies.RenderedVirtualDependency(
                content=content + "\n",
                summary_hash=summary_hash,
                report=scenario.make_report(
                    concrete_annotations={
                        "djangoexample.relations1.models.Abstract": "__virtual__.mod_3327724610.Concrete__Abstract",
                        "djangoexample.relations1.models.Child1": "__virtual__.mod_3327724610.Concrete__Child1",
                        "djangoexample.relations1.models.Child2": "__virtual__.mod_3327724610.Concrete__Child2",
                        "djangoexample.relations1.models.Concrete1": "__virtual__.mod_3327724610.Concrete__Concrete1",
                        "djangoexample.relations1.models.Concrete2": "__virtual__.mod_3327724610.Concrete__Concrete2",
                    },
                    concrete_querysets={
                        "djangoexample.relations1.models.Abstract": "__virtual__.mod_3327724610.ConcreteQuerySet__Abstract",
                        "djangoexample.relations1.models.Child1": "__virtual__.mod_3327724610.ConcreteQuerySet__Child1",
                        "djangoexample.relations1.models.Child2": "__virtual__.mod_3327724610.ConcreteQuerySet__Child2",
                        "djangoexample.relations1.models.Concrete1": "__virtual__.mod_3327724610.ConcreteQuerySet__Concrete1",
                        "djangoexample.relations1.models.Concrete2": "__virtual__.mod_3327724610.ConcreteQuerySet__Concrete2",
                    },
                    report_import_path={
                        "djangoexample.relations1.models": "__virtual__.mod_3327724610"
                    },
                ),
                virtual_import_path=virtual_dependency.summary.virtual_import_path,
            )

            assert hasher_called == [1]

        def test_writing_virtual_dependencies_3(
            self, discovered_django_example: protocols.Discovered[Project]
        ) -> None:
            scenario = self.Scenario(discovered_django_example)

            hasher_called: list[int] = []

            def hasher(*parts: bytes) -> str:
                hasher_called.append(1)
                assert parts == (b"module:djangoexample.empty_models.models",)
                return "__hashed_for_bad__"

            virtual_dependency = scenario.all_virtual_dependencies[
                ImportPath("djangoexample.empty_models.models")
            ]

            content = textwrap.dedent("""
            def interface____differentiated__1() -> None:
                return None

            mod = "djangoexample.empty_models.models"
            summary = "__virtual__.mod_3808300370::djangoexample.empty_models.models::installed_apps=__installed_apps_hash__::significant=__hashed_for_bad__"
            """).strip()

            summary_hash = (
                "__virtual__.mod_3808300370"
                "::djangoexample.empty_models.models"
                "::installed_apps=__installed_apps_hash__"
                "::significant=__hashed_for_bad__"
            )

            written = scenario.scribe(hasher=hasher, virtual_dependency=virtual_dependency)
            assert written == virtual_dependencies.RenderedVirtualDependency(
                content=content + "\n",
                summary_hash=summary_hash,
                report=virtual_dependencies.Report(
                    report_import_path={
                        ImportPath("djangoexample.empty_models.models"): ImportPath(
                            "__virtual__.mod_3808300370"
                        )
                    }
                ),
                virtual_import_path=virtual_dependency.summary.virtual_import_path,
            )

            assert hasher_called == [1]



import os
import pathlib

import pytest

from extended_mypy_django_plugin.django_analysis import ImportPath, protocols, virtual_dependencies


class TestReportInstaller:
    class TestWriteReport:
        def test_can_write_to_the_virtual_import_path(self, tmp_path: pathlib.Path) -> None:
            installer = virtual_dependencies.ReportInstaller(_get_report_summary=lambda path: None)
            assert installer._written == {}

            content = "hello\nthere"
            content2 = "things\nstuff\n"
            content3 = "compelling\nexample\n"

            scratch_root = tmp_path
            assert list(tmp_path.iterdir()) == []

            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="__summary__",
                virtual_import_path=ImportPath("mod_blah"),
                content=content,
            )

            assert len(list(tmp_path.iterdir())) == 1
            location = tmp_path / "mod_blah.py"
            assert location.exists()
            assert location.read_text() == content
            assert installer._written == {location: "__summary__"}

            # With folders
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="__summary2__",
                virtual_import_path=ImportPath("forest.of.trees"),
                content=content2,
            )

            found: list[pathlib.Path] = []
            for root, _, files in os.walk(scratch_root):
                for name in files:
                    found.append(pathlib.Path(root) / name)

            location2 = scratch_root / "forest" / "of" / "trees.py"
            assert found == [location, location2]
            assert location.read_text() == content
            assert location2.read_text() == content2

            assert installer._written == {location: "__summary__", location2: "__summary2__"}

            # And with None summary hash
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash=None,
                virtual_import_path=ImportPath("mod_other"),
                content=content3,
            )

            found = []
            for root, _, files in os.walk(scratch_root):
                for name in files:
                    found.append(pathlib.Path(root) / name)

            location3 = scratch_root / "mod_other.py"
            assert sorted(found) == sorted([location, location3, location2])
            assert location.read_text() == content
            assert location2.read_text() == content2
            assert location3.read_text() == content3

            assert installer._written == {
                location: "__summary__",
                location2: "__summary2__",
                location3: None,
            }

        @pytest.mark.parametrize(
            "bad_path",
            [
                pytest.param("/tmp/other", id="absolute_path"),
                pytest.param("../somewhere", id="relative_path"),
            ],
        )
        def test_complains_if_would_write_outside_scratch_folder(
            self, bad_path: str, tmp_path: pathlib.Path
        ) -> None:
            installer = virtual_dependencies.ReportInstaller(_get_report_summary=lambda path: None)
            assert installer._written == {}

            scratch_root = tmp_path

            with pytest.raises(
                RuntimeError, match="Virtual dependency ends up being outside of the scratch root"
            ):
                installer.write_report(
                    scratch_root=scratch_root,
                    summary_hash="__summary__",
                    virtual_import_path=protocols.ImportPath(bad_path),
                    content="stuff",
                )

    class TestInstallReports:
        def test_it_works_on_empty_folder(self, tmp_path_factory: pytest.TempPathFactory) -> None:
            scratch_root = tmp_path_factory.mktemp("scratch_root")
            destination_holder = tmp_path_factory.mktemp("destination")
            destination = destination_holder / "__virtual__"

            installer = virtual_dependencies.ReportInstaller(_get_report_summary=lambda path: None)
            installer.install_reports(
                scratch_root=scratch_root,
                destination=destination_holder,
                virtual_namespace=ImportPath("__virtual__"),
            )
            assert len(list(destination.iterdir())) == 0

        def test_it_copies_over_everything_that_was_written(
            self, tmp_path_factory: pytest.TempPathFactory
        ) -> None:
            scratch_root = tmp_path_factory.mktemp("scratch_root")
            destination_holder = tmp_path_factory.mktemp("destination")
            destination = destination_holder / "__virtual__"

            installer = virtual_dependencies.ReportInstaller(_get_report_summary=lambda path: None)

            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="s1",
                virtual_import_path=ImportPath("__virtual__.mod_one"),
                content="1",
            )
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="s2",
                virtual_import_path=ImportPath("__virtual__.mod_two"),
                content="2",
            )
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="s3",
                virtual_import_path=ImportPath("__virtual__.mod_three"),
                content="3",
            )

            installer.install_reports(
                scratch_root=scratch_root,
                destination=destination_holder,
                virtual_namespace=ImportPath("__virtual__"),
            )
            assert len(list(destination.iterdir())) == 3

            assert (destination / "mod_one.py").read_text() == "1"
            assert (destination / "mod_two.py").read_text() == "2"
            assert (destination / "mod_three.py").read_text() == "3"

        def test_it_deletes_anything_that_gets_none_summary_and_wasnt_written(
            self, tmp_path_factory: pytest.TempPathFactory
        ) -> None:
            scratch_root = tmp_path_factory.mktemp("scratch_root")
            destination_holder = tmp_path_factory.mktemp("destination")
            destination = destination_holder / "__virtual__"
            destination.mkdir()

            (mod_one := destination / "mod_one.py").write_text("0")
            (mod_two := destination / "mod_two.py").write_text("existing")
            (mod_four := destination / "mod_four.py").write_text("4")
            (mod_five := destination / "mod_five.py").write_text("5")
            (some_dir := destination / "some_dir").mkdir()
            (destination / "some_dir" / "mod_thing.py").write_text("6")
            (destination / "some_dir" / "mod_seven.py").write_text("7")
            (destination / "some_dir" / "subdir").mkdir()
            (destination / "some_dir" / "subdir" / "mod_seven.py").write_text("7")
            nested = destination / "nested"
            hidden = destination / "hidden"
            (hidden / "down").mkdir(parents=True)
            (hidden / "down" / "here.py").write_text("for changing")

            report_summaries = {
                # mod one exists, return a different summary
                mod_one: "different",
                # mod two exists, return the same summary
                mod_two: "s2",
                # mod three doesn't exist, so not called, mod four not written to, has a summary, keep it
                mod_four: "keep",
                # Remove mod_five and everything under some_dir
                # we don't expect _get_report_summary to be called with anything under that directory
                mod_five: None,
                some_dir: None,
                # nested doesn't exist, we want it to not be deleted though
                nested: "_",
                nested / "mc": "_",
                # but hidden does exist and does get compared
                hidden: "_",
                hidden / "down": "_",
                hidden / "down" / "here.py": "_",
            }

            def _get_report_summary(location: pathlib.Path) -> str | None:
                return report_summaries.pop(location)

            installer = virtual_dependencies.ReportInstaller(
                _get_report_summary=_get_report_summary
            )

            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="s1",
                virtual_import_path=ImportPath("__virtual__.mod_one"),
                content="1",
            )
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="s2",
                virtual_import_path=ImportPath("__virtual__.mod_two"),
                content="2",
            )
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="s3",
                virtual_import_path=ImportPath("__virtual__.mod_three"),
                content="3",
            )
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="__nested_hash__",
                virtual_import_path=ImportPath("__virtual__.nested.mc.nestface"),
                content="deep",
            )
            installer.write_report(
                scratch_root=scratch_root,
                summary_hash="__deep_hash__",
                virtual_import_path=ImportPath("__virtual__.hidden.down.here"),
                content="hiding",
            )

            installer.install_reports(
                scratch_root=scratch_root,
                destination=destination_holder,
                virtual_namespace=ImportPath("__virtual__"),
            )
            found: list[pathlib.Path] = []
            for root, _, files in os.walk(destination):
                for name in files:
                    found.append(pathlib.Path(root) / name)

            assert len(found) == 6

            # mod_one was changed
            assert mod_one.read_text() == "1"
            # two existed, had same summary, content doesn't change
            assert (destination / "mod_two.py").read_text() == "existing"
            # Three was added
            assert (destination / "mod_three.py").read_text() == "3"
            # four remained, the rest was deleted
            assert (destination / "mod_four.py").read_text() == "4"
            # our nested import was created with folders
            assert (destination / "nested" / "mc" / "nestface.py").read_text() == "deep"
            # our existing nested import was changed
            assert (destination / "hidden" / "down" / "here.py").read_text() == "hiding"

            # And _get_report_summary was called exactly as many times as our report_summaries had entries before
            assert report_summaries == {}



from .dependency import VirtualDependency, VirtualDependencySummary
from .folder import VirtualDependencyGenerator, VirtualDependencyInstaller
from .handler import VirtualDependencyHandler
from .namer import VirtualDependencyNamer
from .report import (
    CombinedReport,
    RenderedVirtualDependency,
    Report,
    ReportCombiner,
    ReportFactory,
    ReportInstaller,
    ReportSummaryGetter,
    VirtualDependencyScribe,
    make_report_factory,
)

__all__ = [
    "Report",
    "CombinedReport",
    "ReportFactory",
    "ReportCombiner",
    "ReportInstaller",
    "ReportSummaryGetter",
    "make_report_factory",
    "VirtualDependencyNamer",
    "VirtualDependencyScribe",
    "VirtualDependencyHandler",
    "VirtualDependency",
    "VirtualDependencySummary",
    "VirtualDependencyGenerator",
    "VirtualDependencyInstaller",
    "RenderedVirtualDependency",
]



import functools
import os
import pathlib

import pytest

from extended_mypy_django_plugin.django_analysis import (
    ImportPath,
    Project,
    protocols,
    virtual_dependencies,
)
from extended_mypy_django_plugin.version import VERSION

here = pathlib.Path(__file__).parent


def make_report(
    *,
    concrete_annotations: dict[str, str],
    concrete_querysets: dict[str, str],
    report_import_path: dict[str, str],
) -> virtual_dependencies.Report:
    """
    Helper to make the tests below easier to read
    """
    return virtual_dependencies.Report(
        concrete_annotations={
            ImportPath(k): ImportPath(v) for k, v in concrete_annotations.items()
        },
        concrete_querysets={ImportPath(k): ImportPath(v) for k, v in concrete_querysets.items()},
        report_import_path={ImportPath(k): ImportPath(v) for k, v in report_import_path.items()},
    )


class TestEnd2End:
    def test_works(
        self,
        tmp_path_factory: pytest.TempPathFactory,
        discovered_django_example: protocols.Discovered[Project],
    ) -> None:
        count: int = 0

        class VirtualDependencyHandler(
            virtual_dependencies.VirtualDependencyHandler[
                Project,
                virtual_dependencies.VirtualDependency[Project],
                virtual_dependencies.Report,
            ]
        ):
            @classmethod
            def make_project(
                cls, *, project_root: pathlib.Path, django_settings_module: str
            ) -> Project:
                raise NotImplementedError()

            def interface_differentiator(self) -> str:
                nonlocal count
                count += 1
                return f"__differentiated__{count}"

            def get_virtual_namespace(self) -> protocols.ImportPath:
                return ImportPath("__virtual__")

            def hash_installed_apps(self) -> str:
                return "__installed_apps_hash__"

            def make_report_factory(
                self, *, installed_apps_hash: str
            ) -> protocols.ReportFactory[
                virtual_dependencies.VirtualDependency[Project], virtual_dependencies.Report
            ]:
                return virtual_dependencies.make_report_factory(
                    hasher=self.hasher,
                    report_maker=virtual_dependencies.Report,
                    installed_apps_hash=installed_apps_hash,
                    make_differentiator=self.interface_differentiator,
                )

            def virtual_dependency_maker(
                self, *, virtual_dependency_namer: protocols.VirtualDependencyNamer
            ) -> protocols.VirtualDependencyMaker[
                Project, virtual_dependencies.VirtualDependency[Project]
            ]:
                return functools.partial(
                    virtual_dependencies.VirtualDependency.create,
                    discovered_project=self.discovered,
                    virtual_dependency_namer=virtual_dependency_namer,
                )

        destination = tmp_path_factory.mktemp("destination")

        handler = VirtualDependencyHandler(
            discovered=discovered_django_example, hasher=VirtualDependencyHandler.make_hasher()
        )

        report = handler.make_report(virtual_deps_destination=destination)

        assert (
            report.version
            == f"__virtual__|plugin:{VERSION}:installed_apps:__installed_apps_hash__|settings_types:2183014887|written_deps:1749711409"
        )
        assert report.report == make_report(
            concrete_annotations={
                "django.contrib.admin.models.LogEntry": "__virtual__.mod_2456226428.Concrete__LogEntry",
                "django.contrib.auth.models.AbstractUser": "__virtual__.mod_2289830437.Concrete__AbstractUser",
                "django.contrib.auth.models.PermissionsMixin": "__virtual__.mod_2289830437.Concrete__PermissionsMixin",
                "django.contrib.auth.models.Permission": "__virtual__.mod_2289830437.Concrete__Permission",
                "django.contrib.auth.models.Group": "__virtual__.mod_2289830437.Concrete__Group",
                "django.contrib.auth.models.User": "__virtual__.mod_2289830437.Concrete__User",
                "django.contrib.contenttypes.models.ContentType": "__virtual__.mod_3961720227.Concrete__ContentType",
                "django.contrib.sessions.models.Session": "__virtual__.mod_3074165738.Concrete__Session",
                "djangoexample.exampleapp.models.Parent": "__virtual__.mod_3347844205.Concrete__Parent",
                "djangoexample.exampleapp.models.Parent2": "__virtual__.mod_3347844205.Concrete__Parent2",
                "djangoexample.exampleapp.models.Child1": "__virtual__.mod_3347844205.Concrete__Child1",
                "djangoexample.exampleapp.models.Child2": "__virtual__.mod_3347844205.Concrete__Child2",
                "djangoexample.exampleapp.models.Child3": "__virtual__.mod_3347844205.Concrete__Child3",
                "djangoexample.exampleapp.models.Child4": "__virtual__.mod_3347844205.Concrete__Child4",
                "djangoexample.exampleapp2.models.ChildOther": "__virtual__.mod_3537308831.Concrete__ChildOther",
                "djangoexample.exampleapp2.models.ChildOther2": "__virtual__.mod_3537308831.Concrete__ChildOther2",
                "djangoexample.relations1.models.Abstract": "__virtual__.mod_3327724610.Concrete__Abstract",
                "djangoexample.relations1.models.Child1": "__virtual__.mod_3327724610.Concrete__Child1",
                "djangoexample.relations1.models.Child2": "__virtual__.mod_3327724610.Concrete__Child2",
                "djangoexample.relations1.models.Concrete1": "__virtual__.mod_3327724610.Concrete__Concrete1",
                "djangoexample.relations1.models.Concrete2": "__virtual__.mod_3327724610.Concrete__Concrete2",
                "djangoexample.relations2.models.Thing": "__virtual__.mod_3328248899.Concrete__Thing",
                "djangoexample.only_abstract.models.AnAbstract": "__virtual__.mod_4035906997.Concrete__AnAbstract",
                "django.contrib.auth.base_user.AbstractBaseUser": "__virtual__.mod_2833058650.Concrete__AbstractBaseUser",
                "django.contrib.sessions.base_session.AbstractBaseSession": "__virtual__.mod_113708644.Concrete__AbstractBaseSession",
            },
            concrete_querysets={
                "django.contrib.admin.models.LogEntry": "__virtual__.mod_2456226428.ConcreteQuerySet__LogEntry",
                "django.contrib.auth.models.AbstractUser": "__virtual__.mod_2289830437.ConcreteQuerySet__AbstractUser",
                "django.contrib.auth.models.PermissionsMixin": "__virtual__.mod_2289830437.ConcreteQuerySet__PermissionsMixin",
                "django.contrib.auth.models.Permission": "__virtual__.mod_2289830437.ConcreteQuerySet__Permission",
                "django.contrib.auth.models.Group": "__virtual__.mod_2289830437.ConcreteQuerySet__Group",
                "django.contrib.auth.models.User": "__virtual__.mod_2289830437.ConcreteQuerySet__User",
                "django.contrib.contenttypes.models.ContentType": "__virtual__.mod_3961720227.ConcreteQuerySet__ContentType",
                "django.contrib.sessions.models.Session": "__virtual__.mod_3074165738.ConcreteQuerySet__Session",
                "djangoexample.exampleapp.models.Parent": "__virtual__.mod_3347844205.ConcreteQuerySet__Parent",
                "djangoexample.exampleapp.models.Parent2": "__virtual__.mod_3347844205.ConcreteQuerySet__Parent2",
                "djangoexample.exampleapp.models.Child1": "__virtual__.mod_3347844205.ConcreteQuerySet__Child1",
                "djangoexample.exampleapp.models.Child2": "__virtual__.mod_3347844205.ConcreteQuerySet__Child2",
                "djangoexample.exampleapp.models.Child3": "__virtual__.mod_3347844205.ConcreteQuerySet__Child3",
                "djangoexample.exampleapp.models.Child4": "__virtual__.mod_3347844205.ConcreteQuerySet__Child4",
                "djangoexample.exampleapp2.models.ChildOther": "__virtual__.mod_3537308831.ConcreteQuerySet__ChildOther",
                "djangoexample.exampleapp2.models.ChildOther2": "__virtual__.mod_3537308831.ConcreteQuerySet__ChildOther2",
                "djangoexample.relations1.models.Abstract": "__virtual__.mod_3327724610.ConcreteQuerySet__Abstract",
                "djangoexample.relations1.models.Child1": "__virtual__.mod_3327724610.ConcreteQuerySet__Child1",
                "djangoexample.relations1.models.Child2": "__virtual__.mod_3327724610.ConcreteQuerySet__Child2",
                "djangoexample.relations1.models.Concrete1": "__virtual__.mod_3327724610.ConcreteQuerySet__Concrete1",
                "djangoexample.relations1.models.Concrete2": "__virtual__.mod_3327724610.ConcreteQuerySet__Concrete2",
                "djangoexample.relations2.models.Thing": "__virtual__.mod_3328248899.ConcreteQuerySet__Thing",
                "djangoexample.only_abstract.models.AnAbstract": "__virtual__.mod_4035906997.ConcreteQuerySet__AnAbstract",
                "django.contrib.auth.base_user.AbstractBaseUser": "__virtual__.mod_2833058650.ConcreteQuerySet__AbstractBaseUser",
                "django.contrib.sessions.base_session.AbstractBaseSession": "__virtual__.mod_113708644.ConcreteQuerySet__AbstractBaseSession",
            },
            report_import_path={
                "django.contrib.admin.models": "__virtual__.mod_2456226428",
                "django.contrib.auth.models": "__virtual__.mod_2289830437",
                "django.contrib.contenttypes.models": "__virtual__.mod_3961720227",
                "django.contrib.sessions.models": "__virtual__.mod_3074165738",
                "djangoexample.exampleapp.models": "__virtual__.mod_3347844205",
                "djangoexample.exampleapp2.models": "__virtual__.mod_3537308831",
                "djangoexample.relations1.models": "__virtual__.mod_3327724610",
                "djangoexample.relations2.models": "__virtual__.mod_3328248899",
                "djangoexample.only_abstract.models": "__virtual__.mod_4035906997",
                "djangoexample.empty_models.models": "__virtual__.mod_3808300370",
                "django.contrib.auth.base_user": "__virtual__.mod_2833058650",
                "django.contrib.sessions.base_session": "__virtual__.mod_113708644",
            },
        )

        assert len(list(destination.iterdir())) != 0

        found: dict[pathlib.Path, str] = {}
        expected: dict[pathlib.Path, str] = {}

        for path, result in ((destination, found), (here / "generated_reports", expected)):
            for root, _, files in os.walk(path):
                for name in files:
                    location = pathlib.Path(root) / name
                    result[location.relative_to(path)] = location.read_text()

        assert found == expected

        location = destination / handler.get_virtual_namespace() / "mod_3734901629.py"
        assert not location.exists()
        report.ensure_virtual_dependency(
            module_import_path="djangoexample.not_installed_with_concrete.models"
        )
        assert (
            location.read_text()
            == 'mod = "djangoexample.not_installed_with_concrete.models"\nsummary = "||not_installed||"\n'
        )

        assert (
            virtual_dependencies.VirtualDependencyScribe.get_report_summary(location)
            == "||not_installed||"
        )



import inspect
import os
import pathlib
import sys
import textwrap
from collections.abc import Sequence
from typing import TYPE_CHECKING

import pytest

from extended_mypy_django_plugin.django_analysis import ImportPath, project, protocols

project_root = pathlib.Path(__file__).parent.parent.parent


class TestReplacedEnvVarsAndSysPath:
    def test_it_can_handle_env_vars(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("one", "one_val")
        monkeypatch.setenv("two", "two_val")

        monkeypatch.setenv("three", "tobedeleted")
        monkeypatch.delenv("three")

        all_env = dict(os.environ)

        assert "three" not in all_env
        assert all_env["one"] == "one_val"
        assert all_env["two"] == "two_val"

        with project.replaced_env_vars_and_sys_path(
            additional_sys_path=[], env_vars={"one": "blah", "three": "twenty"}
        ):
            changed = dict(os.environ)

        assert dict(os.environ) == all_env
        assert all_env != changed
        assert changed == {**all_env, "one": "blah", "three": "twenty"}

    def test_it_can_handle_changes_to_sys_path(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(sys, "path", ["one", "two", "three"])

        all_path = list(sys.path)
        assert all_path == ["one", "two", "three"]

        with project.replaced_env_vars_and_sys_path(
            additional_sys_path=["two", "four"], env_vars={}
        ):
            changed = list(sys.path)

        assert list(sys.path) == ["one", "two", "three"]
        assert changed == ["one", "two", "three", "four"]


class TestProject:
    def test_getting_an_discovered_project(
        self, pytester: pytest.Pytester, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """
        Using pytester to make it easier to run a test in a subprocess so we don't poison the import space
        """

        def test_getting_project() -> None:
            import dataclasses
            import os
            import pathlib
            from typing import cast

            from django.apps.registry import Apps
            from django.conf import LazySettings

            from extended_mypy_django_plugin.django_analysis import Project

            @dataclasses.dataclass(frozen=True, kw_only=True)
            class FakeModel:
                model_name: str = "MyModel"
                module_import_path: protocols.ImportPath = dataclasses.field(
                    default_factory=lambda: ImportPath("fake.model")
                )
                import_path: protocols.ImportPath = dataclasses.field(
                    default_factory=lambda: ImportPath("fake.model.MyModel")
                )
                is_abstract: bool = False
                default_custom_queryset: protocols.ImportPath | None = None
                all_fields: protocols.FieldsMap = dataclasses.field(default_factory=dict)
                models_in_mro: Sequence[protocols.ImportPath] = dataclasses.field(
                    default_factory=list
                )

            @dataclasses.dataclass(frozen=True, kw_only=True)
            class FakeModule:
                installed: bool = True
                import_path: protocols.ImportPath = dataclasses.field(
                    default_factory=lambda: ImportPath("somewhere")
                )
                defined_models: protocols.ModelMap = dataclasses.field(
                    default_factory=lambda: {fake_model.import_path: fake_model}
                )
                models_hash: str = ""

            fake_model: protocols.Model = FakeModel()
            fake_module: protocols.Module = FakeModule()

            class Discovery:
                def discover_settings_types(
                    self, loaded_project: protocols.Loaded[Project], /
                ) -> protocols.SettingsTypesMap:
                    assert (
                        loaded_project.settings.UNIQUE_SETTING_TO_EXTENDED_MYPY_PLUGIN_DJANGOEXAMPLE  # type: ignore[misc]
                        == "unique"
                    )
                    return {"not": "accurate"}

                def discover_installed_models(
                    self, loaded_project: protocols.Loaded[Project], /
                ) -> protocols.ModelModulesMap:
                    return {fake_module.import_path: fake_module}

                def discover_concrete_models(
                    self, loaded_project: protocols.Loaded[Project], models: protocols.ModelMap, /
                ) -> protocols.ConcreteModelsMap:
                    assert models == {fake_model.import_path: fake_model}
                    return {fake_model.import_path: [fake_model]}

            if TYPE_CHECKING:
                _sta: protocols.Discovery[Project] = cast(Discovery, None)

            root_dir = pathlib.Path(os.environ["PROJECT_ROOT"]) / "example"
            project = Project(
                root_dir=root_dir,
                additional_sys_path=[str(root_dir)],
                discovery=Discovery(),
                env_vars={"DJANGO_SETTINGS_MODULE": "djangoexample.settings"},
            )

            loaded_project = project.load_project()
            discovered_project = loaded_project.perform_discovery()

            assert loaded_project.root_dir == root_dir
            assert loaded_project.env_vars == project.env_vars
            assert isinstance(loaded_project.settings, LazySettings)
            assert (
                loaded_project.settings.UNIQUE_SETTING_TO_EXTENDED_MYPY_PLUGIN_DJANGOEXAMPLE  # type: ignore[misc]
                == "unique"
            )
            assert isinstance(loaded_project.apps, Apps)

            assert discovered_project.loaded_project is loaded_project
            assert discovered_project.installed_apps == [
                "django.contrib.admin",
                "django.contrib.auth",
                "django.contrib.contenttypes",
                "django.contrib.sessions",
                "django.contrib.messages",
                "django.contrib.staticfiles",
                "djangoexample.exampleapp",
                "djangoexample.exampleapp2",
                "djangoexample.only_abstract",
                "djangoexample.no_models",
                "djangoexample.relations1",
                "djangoexample.relations2",
                "djangoexample.empty_models",
            ]
            assert discovered_project.settings_types == {"not": "accurate"}
            assert discovered_project.installed_models_modules == {
                fake_module.import_path: fake_module
            }
            assert discovered_project.all_models == {fake_model.import_path: fake_model}
            assert discovered_project.concrete_models == {fake_model.import_path: [fake_model]}

        test_content = (
            "from extended_mypy_django_plugin.django_analysis import protocols, ImportPath"
            + "\n"
            + "from typing import TYPE_CHECKING"
            + "\n"
            + "from collections.abc import Sequence"
            + "\n\n"
            + textwrap.dedent(inspect.getsource(test_getting_project))
        )
        pytester.makepyfile(test_content)

        monkeypatch.setenv("PROJECT_ROOT", str(project_root))
        result = pytester.runpytest_subprocess("-vvv")
        result.assert_outcomes(passed=1)



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, cast

from django.db import models
from typing_extensions import Self

from extended_mypy_django_plugin.django_analysis import ImportPath, protocols


@dataclasses.dataclass(frozen=True, kw_only=True)
class Field:
    @classmethod
    def create(
        cls, *, model_import_path: protocols.ImportPath, field: protocols.DjangoField
    ) -> Self:
        return cls(
            model_import_path=model_import_path,
            field_type=ImportPath.from_cls(field.__class__),
            related_model=(
                # We only care about related models that are other models that exist
                ImportPath.from_cls(related)
                if (
                    isinstance((related := field.related_model), type)
                    and issubclass(related, models.Model)
                    and not related._meta.auto_created
                )
                else None
            ),
        )

    model_import_path: protocols.ImportPath
    field_type: protocols.ImportPath
    related_model: protocols.ImportPath | None


if TYPE_CHECKING:
    _M: protocols.Field = cast(Field, None)



def interface____differentiated__8() -> None:
    return None

mod = "djangoexample.relations2.models"
summary = "__virtual__.mod_3328248899::djangoexample.relations2.models::installed_apps=__installed_apps_hash__::significant=1229019932"

import django.db.models
import djangoexample.relations2.models
ConcreteQuerySet__Thing = django.db.models.QuerySet[djangoexample.relations2.models.Thing]
Concrete__Thing = djangoexample.relations2.models.Thing



#!/usr/bin/env python
"""
This is used by the Mypy plugin to determine if the Django project has changed in a
way that would require the Mypy Daemon to restart itself.

The django-stubs plugin this relies on loads Django to perform introspection, which means
when Django ORM models or Django settings change, that introspection may no longer be correct.

The nature of how Django relies on globals and caches means it's very difficult to reliably
reload Django in a process where Django has already been loaded, and so this logic must be
done in a separate process.

The script relies on replicating Mypy plugin discovery logic to find a plugin that is based
off `extended_mypy_django_plugin.plugin.PluginProvider` to find the `VirtualDependencyHandler`
object that will determine a "version" for the project. This number is written to a file that
is specified on the command line to communicate back to the existing Mypy daemon process whether
to consider Django as changed or not.
"""

import argparse
import importlib
import os
import pathlib
import re
import sys

from extended_mypy_django_plugin.plugin import ExtraOptions, PluginProvider, protocols


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("--config-file", help="The path to the mypy config file")
    parser.add_argument("--mypy-plugin", action="append", help="The mypy plugins configured")
    parser.add_argument(
        "--version-file", help="File to write the version to", type=pathlib.Path, required=True
    )
    return parser


def main(argv: list[str] | None = None) -> None:
    parser = make_parser()
    args = parser.parse_args(argv)

    # We need to know where the scratch path is so that we can find the special
    # file that tells us to pretend Django hasn't changed
    extra_options = ExtraOptions.from_config(args.config_file)

    if (extra_options.scratch_path / "__assume_django_state_unchanged__").exists():
        # An exit code of 2 tells the existing process to say the version hasn't changed
        sys.exit(2)

    plugin_provider: PluginProvider[protocols.Report] | None = None

    for plugin in args.mypy_plugin:
        found = load_plugin(plugin, args.config_file)
        if isinstance(found, PluginProvider):
            plugin_provider = found
            break

    if plugin_provider is None:
        raise ValueError("Couldn't find the extension that provides extended_mypy_django_plugin")

    report = plugin_provider.plugin_cls.make_virtual_dependency_report(
        extra_options=extra_options,
        virtual_dependency_handler=plugin_provider.virtual_dependency_handler,
    )

    # Communicate back the version we have. This is done explicitly rather than
    # by relying on filtering stdout or stderr
    args.version_file.write_text(report.version)


def load_plugin(plugin_path: str, config_file: str) -> object | None:
    """
    This is heavily based off what mypy itself does to load plugins
    """
    func_name = "plugin"
    plugin_dir: str | None = None
    if ":" in os.path.basename(plugin_path):
        plugin_path, func_name = plugin_path.rsplit(":", 1)

    if plugin_path.endswith(".py"):
        # Plugin paths can be relative to the config file location.
        plugin_path = os.path.join(os.path.dirname(config_file), plugin_path)
        if not os.path.isfile(plugin_path):
            return None

        plugin_dir = os.path.abspath(os.path.dirname(plugin_path))
        fnam = os.path.basename(plugin_path)
        module_name = fnam[:-3]
        sys.path.insert(0, plugin_dir)
    elif re.search(r"[\\/]", plugin_path):
        # Plugin does not have a .py extension
        return None
    else:
        module_name = plugin_path

    try:
        module = importlib.import_module(module_name)
    except Exception:
        return None
    finally:
        if plugin_dir is not None:
            assert sys.path[0] == plugin_dir
            del sys.path[0]

    if not hasattr(module, func_name):
        return None

    return getattr(module, func_name, None)


if __name__ == "__main__":
    main()



def interface____differentiated__11() -> None:
    return None

mod = "django.contrib.auth.base_user"
summary = "__virtual__.mod_2833058650::django.contrib.auth.base_user::installed_apps=__installed_apps_hash__::significant=3626250221"

import django.contrib.auth.base_user
import django.contrib.auth.models
import django.db.models
ConcreteQuerySet__AbstractBaseUser = django.db.models.QuerySet[django.contrib.auth.models.User]
Concrete__AbstractBaseUser = django.contrib.auth.models.User
